"""考试相关的模块
"""
from __future__ import annotations

from typing import Dict, Any, Optional

import requests

from .._lib_wrapper.encrypt import pad16, aes_ecb_encryptor

from ..utils.request_transformer import Request, RequestTransformer

__all__ = ['get_exam_raw', 'async_get_exam_raw']

__exam_encryptor = aes_ecb_encryptor("cquisse123456789".encode())
EXAM_LIST_URL = "https://my.cqu.edu.cn/api/exam/examTask/get-student-exam-tab-list"


@RequestTransformer.register()
def _get_exam_raw(session: Request, student_id: str) -> Dict[str, Any]:
    return (yield session.get(EXAM_LIST_URL,
                              params={"studentId": __exam_encryptor(pad16(student_id.encode())).hex().upper()}
                              )).json()

def get_exam_raw(student_id: str, session: Optional[requests.Session] = None) -> Dict[str, Any]:
    """获取考表的原始 json 数据（被反序列化为 python 字典对象）

    :param student_id: 学号
    :type student_id: str
    :param session: 用于请求的 requests session
    :type session: requests.Session, optional
    :return: 反序列化后的课表 json 数据
    :rtype: Dict[str, Any]
    """
    return _get_exam_raw.sync_request((session or requests), student_id)

async def async_get_exam_raw(session: Request, student_id: str) -> Dict[str, Any]:
    """获取考表的原始 json 数据（被反序列化为 python 字典对象）

    :param session: 用于请求的 requests session
    :type session: requests.Session
    :param student_id: 学号
    :type student_id: str
    :return: 反序列化后的课表 json 数据
    :rtype: Dict[str, Any]
    """
    return await _get_exam_raw.async_request((session or requests), student_id)



from __future__ import annotations

from typing import Generic, Optional

from requests import Session

from ...exception import MycquUnauthorized
from ...utils.request_transformer import Request
from pydantic import BaseModel

__all__ = ["User",]


class User(BaseModel, Generic[Request]):
    """用户信息"""

    name: str
    """姓名"""
    id: str
    """统一身份认证号"""
    code: str
    """学工号"""
    role: str
    """身份，已知取值有学生 :obj:`"student"`、教师 :obj:`"instructor`"`"""
    email: Optional[str]
    "电子邮箱"
    phone_number: Optional[str]
    "电话号码"

    @staticmethod
    def fetch_self(session: Session) -> User:
        """从在 mycqu 认证了的会话获取当前登录用户的信息

        :param session: 登陆了统一身份认证的会话
        :type session: Session
        :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
        :return: 当前用户信息
        :rtype: User
        """
        resp = session.get("https://my.cqu.edu.cn/authserver/simple-user")
        if resp.status_code == 401:
            raise MycquUnauthorized()
        data = resp.json()
        return User(
            name=data["name"],
            code=data["code"],
            id=data["username"],
            role=data["type"],
            email=data["email"],
            phone_number=data["phoneNumber"]
        )

    @staticmethod
    async def async_fetch_self(session: Request) -> User:
        """
        异步的从在 mycqu 认证了的会话获取当前登录用户的信息

        :param session: 登陆了统一身份认证的会话
        :type session: Session
        :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
        :return: 当前用户信息
        :rtype: User
        """
        resp = await session.get("https://my.cqu.edu.cn/authserver/simple-user")
        if resp.status_code == 401:
            raise MycquUnauthorized()
        data = resp.json()
        return User(
            name=data["name"],
            code=data["code"],
            id=data["username"],
            role=data["type"],
            email=data["email"],
            phone_number=data["phoneNumber"]
        )


from functools import wraps, partial
from inspect import isgeneratorfunction, isgenerator
from typing import Callable, Any, Generator, Tuple

from ..config import ConfigManager
from .models import RequestReturns, RequestProtocol, Requestable, Request, Response


__all__ = ['RequestTransformer', ]


class RequestTransformer:
    def __init__(self, generator: Callable[..., Generator[RequestReturns, Any, Any]]):
        """
        拓展按照一定格式书写的生成器函数以同时支持同步/异步发出请求
        直接调用该类实例则默认以同步的方式执行此函数

        :param sync_request_param_mapper: 用于发出同步请求时使用的参数转换库，默认为`RequestsParamsMapper`
        :param async_request_param_mapper: 用于发出异步请求时使用的参数转换库，默认为`HttpxParamsMapper`
        """
        if not (isgeneratorfunction(generator) or isgenerator(generator)):
            self.without_request = True
        else:
            self.without_request = False
        self.generator = generator
        self.instance = None

    def __get__(self, instance, owner):
        self.instance = instance if instance is not None else owner
        return self


    @property
    def sync_request(self) -> Callable[[RequestProtocol, Any, ...], Any]:
        """
        将生成器函数以同步的方式执行，返回同步函数
        """
        if self.without_request:
            if self.instance is not None:
                return partial(self.generator, self.instance)
            else:
                return self.generator

        @wraps(self.generator)
        def inner_function(request: Request, *args, **kwargs) -> Response:
            try:
                if self.instance is not None:
                    generator = self.generator(self.instance, Requestable(request), *args, **kwargs)
                else:
                    generator = self.generator(Requestable(request), *args, **kwargs)
                res = None
                while True:
                    request_returns: RequestReturns = generator.send(res)
                    if isinstance(request_returns, RequestTransformer):
                        request_returns = (request_returns, {})
                    if isinstance(request_returns[0], RequestTransformer):
                        res = request_returns[0].sync_request(request, **request_returns[1])
                    else:
                        res = request_returns[0].request(
                            **request_returns[1].to_param_dict(
                                ConfigManager().config['request']['sync_request_params_mapper']
                            )
                        )
            except StopIteration as e:
                return e.value

        return inner_function

    @property
    def async_request(self) -> Callable[[RequestProtocol, Any, ...], Any]:
        """
        将生成器函数以异步的方式执行，返回async function
        """
        if self.without_request:
            @wraps(self.generator)
            async def return_function(*args, **kwargs):
                if self.instance is not None:
                    return self.generator(self.instance, *args, **kwargs)
                else:
                    return self.generator(*args, **kwargs)
            return return_function
        @wraps(self.generator)
        async def inner_function(request: Request, *args, **kwargs) -> Response:
            try:
                if self.instance is not None:
                    generator = self.generator(self.instance, Requestable(request), *args, **kwargs)
                else:
                    generator = self.generator(Requestable(request), *args, **kwargs)
                res = None
                while True:
                    request_returns: RequestReturns = generator.send(res)
                    if isinstance(request_returns, RequestTransformer):
                        request_returns = (request_returns, {})
                    if isinstance(request_returns[0], RequestTransformer):
                        res = await request_returns[0].async_request(request, **request_returns[1])
                    else:
                        res = await request_returns[0].request(
                            **request_returns[1].to_param_dict(
                                ConfigManager().config['request']['async_request_params_mapper']
                            )
                        )
            except StopIteration as e:
                return e.value

        return inner_function


    @classmethod
    def register(cls):
        """
        将按照一定格式书写的同步函数包装成`RequestTransformer`对象
        """
        def wrapped_function(func: Callable[..., Generator[RequestReturns, Any, Any]]):
            return cls(func)

        return wrapped_function



from __future__ import annotations

from typing import Any, Dict, Tuple, List
from pydantic import BaseModel

from ...utils.datetimes import parse_period_str, parse_weeks_str
from ...utils.period import Period

__all__ = ['RoomActivityInfo']


class RoomActivityInfo(BaseModel):
    """教室活动的公有属性"""
    period: Period
    """占用节数"""
    weeks: List[Period]
    """行课周数，列表中每个元组 (a,b) 代表一个周数范围 a~b（包含 a, b），在单独的一周则有 b=a"""
    weekday: int
    """星期，0 为周一，6 为周日，此与 :attr:`datetime.date.day` 一致"""

    @staticmethod
    def from_dict(data: Dict[str, Any]):
        """从反序列化的一个活动信息 json 中生成RoomActivityInfo对象

        :param data: 反序列化成字典的活动 json
        :type data: Dict[str, Any]
        :return: 教室活动
        :rtype: RoomActivityInfo
        """
        return RoomActivityInfo(
            period=parse_period_str(data['periodFormat']),
            weeks=parse_weeks_str(data['teachingWeekFormat']),
            weekday=int(data['weekDay']) - 1
        )


from __future__ import annotations

import json
from typing import Any, Dict, Optional, List, Tuple

from pydantic import BaseModel
from requests import Session
from datetime import date, datetime

from ...utils.request_transformer import RequestTransformer, Request
from ..tools import parse_response, _get_borrow_books_raw, _renew_book
from ...utils.datetimes import date_from_str, datetime_from_str

__all__ = ['BookInfo']


class BookInfo(BaseModel):
    """
    图书馆书籍相关信息
    """
    id: Optional[int] = None
    """书籍id"""
    title: str
    """书籍名称"""
    call_no: str
    """书籍检索号"""
    library_name: str
    """所属图书馆（如虎溪图书馆自然科学阅览室等）"""
    borrow_time: datetime
    """借出时间"""
    should_return_time: Optional[date] = None
    """应归还日期"""
    is_return: bool
    """是否被归还"""
    return_time: Optional[date] = None
    """归还时间"""
    renew_count: int
    """续借次数"""
    can_renew: bool
    """是否可被续借"""

    @staticmethod
    def from_list(*data: List[str], is_curr: bool) -> List[BookInfo]:
        max_len = max(map(lambda x: len(x), data))
        standard_data = list(map(lambda x: x if len(x) == max_len else [None for i in range(max_len)], data))
        book_infos: List[Tuple] = list(zip(*standard_data))

        result: List[BookInfo] = []

        for book_info in book_infos:
            result.append(
                BookInfo(
                    id=book_info[0],
                    title=book_info[1],
                    call_no=book_info[2],
                    library_name=book_info[3],
                    borrow_time=datetime_from_str(book_info[4]),
                    should_return_time=date_from_str(book_info[5]) if book_info[5] is not None else None,
                    is_return= not is_curr,
                    return_time=date_from_str(book_info[6]) if book_info[6] is not None else None,
                    renew_count=book_info[7],
                    can_renew=book_info[8] if book_info[8] is not None else False,
                )
            )

        return result

    @staticmethod
    @RequestTransformer.register()
    def _fetch(session: Request, is_curr: bool) -> List[BookInfo]:
        res = yield _get_borrow_books_raw, {'is_curr': is_curr}
        data = res.get('columns')
        if data is None:
            return []

        ids = parse_response(data, 'bookId')
        titles = parse_response(data, 'title')
        call_nos = parse_response(data, 'indexNumber')
        library_names = parse_response(data, 'roomName')
        borrow_times = parse_response(data, 'borrowDate')
        should_return_times = parse_response(data, 'shouldReturnDate')
        return_times = parse_response(data, 'returnDate')
        renew_count = parse_response(data, 'renewalNumber')
        can_renews = parse_response(data, 'renewflag')

        return BookInfo.from_list(ids, titles, call_nos, library_names,
                                  borrow_times, should_return_times,
                                  return_times, renew_count, can_renews,
                                  is_curr=is_curr)


    @staticmethod
    def fetch(session: Session, is_curr: bool) -> List[BookInfo]:
        """
        获取当前/历史借阅书籍

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu和lib.cqu.edu.cn 进行了认证（:func:`.mycqu.access_mycqu` :func:`.library.access_library`）的 requests 会话
        :type session: Session
        :param is_curr: 是否获取当前借阅书籍（为否则获取历史借阅书籍）
        :type is_curr: bool
        :return: 图书对象组成的列表
        :rtype: List[BookInfo]
        """
        return BookInfo._fetch.sync_request(session, is_curr)

    @staticmethod
    async def async_fetch(session: Session, is_curr: bool) -> List[BookInfo]:
        """
        异步的获取当前/历史借阅书籍

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu和lib.cqu.edu.cn 进行了认证（:func:`.mycqu.access_mycqu` :func:`.library.access_library`）的 requests 会话
        :type session: Session
        :param is_curr: 是否获取当前借阅书籍（为否则获取历史借阅书籍）
        :type is_curr: bool
        :return: 图书对象组成的列表
        :rtype: List[BookInfo]
        """
        return await BookInfo._fetch.async_request(session, is_curr)

    @staticmethod
    def _parse_renew_result(res: Dict) -> str:
        if res.get('data') is not None:
            if res['data'].get('code') == 200:
                return "success"
            elif res['data'].get('message') is not None and res['data']['message'] != "":
                return res['data']['message']
        return "未知异常"

    @staticmethod
    def renew(session: Session, book_id: str) -> str:
        return BookInfo._parse_renew_result(_renew_book.sync_request(session, book_id))


    @staticmethod
    async def async_renew(session: Request, book_id: str) -> str:
        return BookInfo._parse_renew_result((await _renew_book.async_request(session, book_id)))



from .cqu_session import CQUSession
from .cqu_session_info import CQUSessionInfo
from .course import Course
from .course_day_time import CourseDayTime
from .course_timetable import CourseTimetable

__all__ = ['Course', 'CourseDayTime', 'CourseTimetable', 'CQUSession', 'CQUSessionInfo']



from typing import Callable
from .utils.request_transformer.models import Response
__all__ = ["MycquException", "CQUWebsiteError", "NotAllowedService", "NeedCaptcha", "InvaildCaptcha",
           "IncorrectLoginCredentials", "TicketGetError", "ParseError", "MycquUnauthorized",
           "UnknownAuthserverException", "NotLogined", "MultiSessionConflict", "InvalidRoom", "CQUSessionIdNotExist"]

class MycquException(Exception):
    pass

class CQUWebsiteError(MycquException):
    """重大网站返回未知的错误或异常时抛出"""

    def __init__(self, error_msg='No error info'):
        super().__init__('CQU website return error: ' + error_msg)


class NotAllowedService(MycquException):
    """试图认证不允许的服务时抛出
    """


class NeedCaptcha(MycquException):
    """登录统一身份认证时需要输入验证码时拋出
    """

    def __init__(self, image: bytes, image_type: str, after_captcha: Callable[[str], Response]):
        super().__init__("captcha is needed")
        self.image: bytes = image
        """验证码图片文件数据"""
        self.image_type: str = image_type
        """验证码图片 MIME 类型"""
        self.after_captcha: Callable[[str], Response] = after_captcha
        """将验证码传入，调用以继续进行登陆"""


class InvaildCaptcha(MycquException):
    """登录统一身份认证输入了无效验证码时抛出
    """

    def __init__(self):
        super().__init__("invaild captcha")


class IncorrectLoginCredentials(MycquException):
    """使用无效的的登录凭据（如错误的用户、密码）
    """

    def __init__(self):
        super().__init__("incorrect username or password")


class TicketGetError(CQUWebsiteError):
    """
    当未能从网页对应位置中获取到ticket时抛出
    """


class ParseError(CQUWebsiteError):
    """
    当从返回数据解析所需值失败时抛出
    """


class UnknownAuthserverException(CQUWebsiteError):
    """登录或认证服务过程中未知错误"""


class NotLogined(MycquException):
    """未登陆或登陆过期的会话被用于进行需要统一身份认证登陆的操作
    """

    def __init__(self):
        super().__init__("not in logined status")


class MultiSessionConflict(MycquException):
    """当前用户启用单处登录，并且存在其他登录会话时抛出"""

    def __init__(self, kick: Callable[[], Response], cancel: Callable[[], Response]):
        super().__init__("单处登录 enabled, kick other sessions of the user or cancel")
        self.kick: Callable[[], Response] = kick
        """踢掉其他会话并登录"""
        self.cancel: Callable[[], Response] = cancel
        """取消登录"""


class MycquUnauthorized(MycquException):
    """访问需要 mycqu 认证，但未对 my.cqu.edu.cn 进行认证或认证过期时抛出"""

    def __init__(self):
        super().__init__("Unauthorized in mycqu, auth.login firstly and then mycqu.access_mycqu")

class InvalidRoom(MycquException):

    def __init__(self):
        super().__init__("Invalid Room Name")


class CQUSessionIdNotExist(MycquException):
    """CQUSession不存在对应Session ID"""
    def __init__(self):
        super().__init__("CQUSession不存在对应Session ID")



"""图书馆相关模块"""

from __future__ import annotations

from typing import Any, Dict, List

from requests import Session

from ..auth import access_service, async_access_service
from ..utils.request_transformer import RequestTransformer, Request

__all__ = ['access_library', 'async_access_library',
           'get_curr_books_raw', 'get_history_books_raw', 'async_get_borrow_books_raw',
           'parse_response']

ACCESS_URL = "http://lib.cqu.edu.cn:8002/api/Auth/AccessToken"
AUTHORIZE_URL = "http://lib.cqu.edu.cn:8000/useridentify/api/third-part-auth/token-by-cas-for-verify-first-login"
CURR_BOOKS_URL = "http://lib.cqu.edu.cn:8000/opac/api/user-opac-center/loan-list"
HISTORY_BOOKS_URL = "http://lib.cqu.edu.cn:8000/opac/api/user-opac-center/history-loan-list"
BOOK_SEARCH_URL = "http://lib.cqu.edu.cn:8000/articlesearch/api/search/asset-search"
RENEW_BOOK_URL = "http://lib.cqu.edu.cn:8000/opac/api/user-opac-center/renew"
GET_BOOK_POS_URL = "http://lib.cqu.edu.cn:8000/opac/api/search-collection-status/collection-status-by-id"


@RequestTransformer.register()
def _access_library(request: Request, ticket: str) -> str:
    res1 = yield request.post(ACCESS_URL, json={
        "orgCode": "cqu",
        "OrgTokenLink": ACCESS_URL
    })
    token1 = res1.json()["data"]["token"]
    res2 = yield request.get(
        AUTHORIZE_URL,
        params={'service': "http://lib.cqu.edu.cn/", "ticket": ticket}, headers={'Authorization': "Bearer " + token1})
    token2 = res2.json()["data"]["token"]
    return "Bearer " + token2



def access_library(session: Session):
    """
    通过统一身份认证登陆图书馆页面，返回UserID和UserKey用于查询

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
    :type session: Session
    """
    res = access_service(session, "http://lib.cqu.edu.cn/")
    ticket = res.url[30:]
    token = _access_library.sync_request(session, ticket)
    session.headers["Authorization"] = token

async def async_access_library(session: Request):
    """
    异步的通过统一身份认证登陆图书馆页面，返回UserID和UserKey用于查询

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
    :type session: Session
    """
    res = await async_access_service(session, "http://lib.cqu.edu.cn/")
    ticket = str(res.url)[30:]
    token = await _access_library.async_request(session, ticket)
    session.headers["Authorization"] = token


@RequestTransformer.register()
def _get_borrow_books_raw(request: Request, is_curr: bool):
    res = yield request.get(CURR_BOOKS_URL if is_curr else HISTORY_BOOKS_URL)
    return res.json()['data']['data']


def get_curr_books_raw(session: Session) -> Dict:
    """
    获取当前借阅书籍

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
    :type session: Session
    :return: 反序列化的书籍信息json
    :rtype: Dict
    """
    return _get_borrow_books_raw.sync_request(session, True)


def get_history_books_raw(session: Session) -> Dict:
    """
    获取历史借阅书籍

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
    :type session: Session
    :return: 反序列化的书籍信息json
    :rtype: Dict
    """
    return _get_borrow_books_raw.sync_request(session, False)

async def async_get_borrow_books_raw(session: Request, is_curr: bool) -> Dict:
    """
    异步的获取当前书籍借阅信息

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
    :type session: Session
    :param is_curr: 是否为当前借阅信息（为False回传历史借阅信息
    :return: 反序列化的书籍信息json
    :rtype: Dict
    """
    return await _get_borrow_books_raw.async_request(session, is_curr)

@RequestTransformer.register()
def _renew_book(request: Request, book_id: int) -> Dict:
    res = yield request.post(RENEW_BOOK_URL, json={
        'data': str(book_id)
    })

    return res.json()

@RequestTransformer.register()
def _get_book_pos(request: Request, book_id: int):
    res = yield request.get(params={'bookid': book_id})
    return res.json().get('data')

def parse_response(target: List[Dict], field_name: str) -> List[str]:
    result = list(filter(lambda x: x['fieldName'] == field_name, target))
    return result[0].get('values') if len(result) != 0 else []



from enum import Enum
from typing import Protocol


__all__ = ['RequestParamsMapper', 'RequestsParamsMapper', 'HttpxParamsMapper']


class RequestParamsMapper(Protocol):
    """
    至少具有method、url、allow_redirects属性的对象

    `RequestParams`的to_param_dict方法基于满足此协议的枚举生成请求参数字典
    `RequestParams`对象传入的参数应当在对应枚举类中全部声明
    可以参考`RequestsParamsMapper`的实现
    """
    @property
    def method(self) -> Enum: ...

    @property
    def url(self) -> Enum: ...

    @property
    def allow_redirects(self) -> Enum: ...


class RequestsParamsMapper(Enum):
    """
    适用于requests库的ParamsMapper
    """
    method = 'method'
    url = 'url'
    params = 'params'
    data = 'data'
    json = 'json'
    headers = 'headers'
    cookies = 'cookies'
    timeout = 'timeout'
    allow_redirects = 'allow_redirects'

class HttpxParamsMapper(Enum):
    """
    适用于httpx库的ParamsMapper
    """
    method = 'method'
    url = 'url'
    params = 'params'
    data = 'data'
    json = 'json'
    headers = 'headers'
    cookies = 'cookies'
    timeout = 'timeout'
    allow_redirects = 'follow_redirects'


from __future__ import annotations

from typing import List, Dict, Any, Optional

from pydantic import BaseModel

from ..tools import get_enroll_list_raw, async_get_enroll_list_raw
from ...course import Course
from ...utils.request_transformer import Request

from requests import Session

__all__ = ['EnrollCourseInfo']

class EnrollCourseInfo(BaseModel):
    """
    可选课程信息
    """
    id: str
    """可选课程id"""
    course: Course
    """可选课程信息"""
    category: str
    """可选课程类型，如：公共基础课，主修专业课，非限制选修课等"""
    type: str
    """课程类别，如：主修专业课，通识教育课程等"""
    enroll_sign: Optional[str] = None
    """选课标识，如：已选，已选满等，当为 :obj:`None` 时标识无相关标记"""
    course_nature: str
    """课程属性，如必修，选修等"""
    campus: List[str]
    """可选课程可选校区，如['D区'], ['A区', 'D区']等"""

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> EnrollCourseInfo:
        """
        从反序列化的选课列表中返回``EnrollCourseInfo``
        :param data: 反序列化成字典的选课列表
        :type data: Dict[str, Any]
        :return: 对应的可选课程列表
        :rtype: EnrollCourseInfo
        """
        return EnrollCourseInfo(
            id=data['id'],
            course=Course(name=data['name'], code=data['codeR'], dept=data['departmentName'],
                          credit=float(data['credit']), course_num=None, instructor=None, session=None),
            category=data['courseCategory'],
            type=data['selectionArea'],
            enroll_sign=data['courseEnrollSign'],
            course_nature=data['courseNature'],
            campus=data['campusShortNameSet']
        )

    @staticmethod
    def fetch(session: Session, is_major: bool = True) -> Dict[str, list[EnrollCourseInfo]]:
        """从 my.cqu.edu.cn 上获取学生可选课程

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :param is_major: 是否获取主修可选课程，为`False`时查询辅修可选课程
        :type is_major: bool
        :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
        :return: 获取的可选课程对象的列表
        :rtype: List[EnrollCourseInfo]
        """
        res = get_enroll_list_raw(session, is_major)
        result = {}
        for key, value in res.items():
            result[key] = [EnrollCourseInfo.from_dict(item) for item in value]

        return result

    @staticmethod
    async def async_fetch(session: Request, is_major: bool = True) -> Dict[str, list[EnrollCourseInfo]]:
        """
        异步的从 my.cqu.edu.cn 上获取学生可选课程

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :param is_major: 是否获取主修可选课程，为`False`时查询辅修可选课程
        :type is_major: bool
        :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
        :return: 获取的可选课程对象的列表
        :rtype: List[EnrollCourseInfo]
        """
        res = await async_get_enroll_list_raw(session, is_major)
        result = {}
        for key, value in res.items():
            result[key] = [EnrollCourseInfo.from_dict(item) for item in value]

        return result



import json
import datetime

from requests import Session

from ._help import _get_ticket, _get_synjones_auth, _get_fee_data, _CardPageParser, _get_hall_ticket
from ..exception import CQUWebsiteError
from ..auth import access_service, async_access_service
from ..utils.datetimes import TIMEZONE
from ..utils.request_transformer import Request, RequestTransformer

# 缴费大厅页面的不同缴费项目的id不同，虎溪和老校区不同
FEE_ITEM_ID = {'Huxi': '182',
               'Old': '181'}

LOGIN_URL = 'http://card.cqu.edu.cn:7280/ias/prelogin?sysid=FWDT'

__all__ = ['get_fees_raw', 'get_card_raw', 'get_bill_raw', 'access_card',
           'async_get_fees_raw', 'async_get_card_raw', 'async_get_bill_raw', 'async_access_card']


def access_card(session: Request):
    """用登陆了统一身份认证的会话在 card.cqu.edu.cn 进行认证

    :param session: 登陆了统一身份认证的会话
    :type session: Session
    """
    res = access_service(session, LOGIN_URL)
    res = session.get(res.headers["Location"])

    parser = _CardPageParser()
    parser.feed(res.text)
    ssoticket_id = parser.ssoticket_id
    _get_hall_ticket.sync_request(session, ssoticket_id)

async def async_access_card(session: Request):
    """
    异步的用登陆了统一身份认证的会话在 card.cqu.edu.cn 进行认证

    :param session: 登陆了统一身份认证的会话
    :type session: Session
    """
    res = await async_access_service(session, LOGIN_URL)
    res = await session.get(res.headers["Location"])

    parser = _CardPageParser()
    parser.feed(res.text)
    ssoticket_id = parser.ssoticket_id
    await _get_hall_ticket.async_request(session, ssoticket_id)

@RequestTransformer.register()
def _get_fees_raw(session: Request, is_huxi: bool, room: str):
    ticket = yield _get_ticket
    synjones_auth = yield _get_synjones_auth, {'ticket': ticket}
    return (yield _get_fee_data, {'synjones_auth': synjones_auth, 'room': room,
                                  'fee_item_id': FEE_ITEM_ID['Huxi'] if is_huxi else FEE_ITEM_ID['Old']})

def get_fees_raw(session: Session, is_huxi: bool, room: str):
    """
    从card.cqu.edu.cn获取水电费详情

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 card.cqu.edu.cn 进行了认证（:func:`.card.access_card`）的 requests 会话
    :type session: Session
    :param is_huxi: 房间号是否为虎溪校区的房间
    :type is_huxi: bool
    :param room: 需要获取水电费详情的宿舍
    :type room: str
    :raises NetworkError: 当访问相关网页时statue code不为200时抛出
    :raises TicketGetError: 当未能从网页对应位置中获取到ticket时抛出
    :raises ParseError: 当从返回数据解析所需值失败时抛出
    :raises CQUWebsiteError: 当网页获取水电费状态码不为success时抛出
    :return: 反序列化获取水电费信息的json
    :rtype: dict
    """
    return _get_fees_raw.sync_request(session, is_huxi, room)

async def async_get_fees_raw(session: Request, is_huxi: bool, room: str):
    """
    异步的从card.cqu.edu.cn获取水电费详情

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 card.cqu.edu.cn 进行了认证（:func:`.card.access_card`）的 requests 会话
    :type session: Session
    :param is_huxi: 房间号是否为虎溪校区的房间
    :type is_huxi: bool
    :param room: 需要获取水电费详情的宿舍
    :type room: str
    :raises NetworkError: 当访问相关网页时statue code不为200时抛出
    :raises TicketGetError: 当未能从网页对应位置中获取到ticket时抛出
    :raises ParseError: 当从返回数据解析所需值失败时抛出
    :raises CQUWebsiteError: 当网页获取水电费状态码不为success时抛出
    :return: 反序列化获取水电费信息的json
    :rtype: dict
    """
    return await _get_fees_raw.async_request(session, is_huxi, room)

@RequestTransformer.register()
def _get_card_raw(session: Request):
    url = "http://card.cqu.edu.cn/NcAccType/GetCurrentAccountList"

    res = yield session.post(url)
    result = json.loads(json.loads(res.text))
    if result['respCode'] != "0000":
        raise CQUWebsiteError(error_msg=result['respInfo'])

    return result['objs'][0]

def get_card_raw(session: Session):
    """
    从card.cqu.edu.cn获取校园卡信息

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 card.cqu.edu.cn 进行了认证（:func:`.card.access_card`）的 requests 会话
    :type session: Session
    :raises CQUWebsiteError: 当网页获取状态码不为0000时抛出
    :return: 获取的校园卡信息
    :rtype: dict
    """
    return _get_card_raw.sync_request(session)

async def async_get_card_raw(session: Session):
    """
    异步的从card.cqu.edu.cn获取校园卡信息

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 card.cqu.edu.cn 进行了认证（:func:`.card.access_card`）的 requests 会话
    :type session: Session
    :raises CQUWebsiteError: 当网页获取状态码不为0000时抛出
    :return: 获取的校园卡信息
    :rtype: dict
    """
    return await _get_card_raw.async_request(session)

@RequestTransformer.register()
def _get_bill_raw(session: Session, account: str, duration: int):
    url = 'http://card.cqu.edu.cn/NcReport/GetMyBill'

    end_date = datetime.datetime.now(tz=TIMEZONE)
    start_date = end_date - datetime.timedelta(duration)

    data = {
        'sdate': start_date.strftime('%Y-%m-%d'),
        'edate': end_date.strftime('%Y-%m-%d'),
        'account': account,
        'page': 1,
        'row': 100,
    }

    res = yield session.post(url=url, data=data)
    result = json.loads(res.text)

    return result['rows']

def get_bill_raw(session: Session, account: str, duration: int):
    """
    从card.cqu.edu.cn获取校园卡账单

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 card.cqu.edu.cn 进行了认证（:func:`.card.access_card`）的 requests 会话
    :type session: Session
    :param account: 校园卡卡号
    :type account: int
    :param duration: 查询时间间隔(默认为30天)
    :type duration: int
    :return: 获取的校园卡账单信息
    :rtype: dict
    """
    return _get_bill_raw.sync_request(session, account, duration)

async def async_get_bill_raw(session: Session, account: str, duration: int):
    """
    异步的从card.cqu.edu.cn获取校园卡账单

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 card.cqu.edu.cn 进行了认证（:func:`.card.access_card`）的 requests 会话
    :type session: Session
    :param account: 校园卡卡号
    :type account: int
    :param duration: 查询时间间隔(默认为30天)
    :type duration: int
    :return: 获取的校园卡账单信息
    :rtype: dict
    """
    return await _get_bill_raw.async_request(session, account, duration)



from __future__ import annotations

from typing import Dict, Optional

from pydantic import BaseModel


__all__ = ['Invigilator']

class Invigilator(BaseModel):
    """监考员信息
    """
    name: str
    """监考员姓名"""
    dept: str
    """监考员所在学院（可能是简称，如 :obj:`"数统"`）"""

    @staticmethod
    def from_dict(data: Dict[str, Optional[str]]) -> Invigilator:
        """从反序列化后的 json 数据中一名正/副监考员的数据中生成 :class:`Invigilator` 对象。

        :param data: 反序列化后的 json 数据中的一次考试数据
        :type data: Dict[str, Optional[str]]
        :return: 对应的 :class:`Invigilator` 对象
        :rtype: Invigilator
        """
        return Invigilator(
            name=data["instructor"],  # type: ignore
            dept=data["instDeptShortName"]  # type: ignore
        )


from typing import List, Dict, Any

from datetime import date

from pydantic import BaseModel

from .room_activity_info import RoomActivityInfo
from ...utils.datetimes import date_from_str

__all__ = ['RoomTempActivity']


class RoomTempActivity(BaseModel):
    """教室临时活动"""
    activity_info: RoomActivityInfo
    """基础信息"""
    content: str
    """临时活动名称"""
    department: str
    """申请学院"""
    type: str
    """临时活动类型，如上课、开会等"""
    time_range: str
    """活动持续时间"""
    date: List[date]
    """活动举办日期"""

    @staticmethod
    def from_dict(data: Dict[str, Any]):
        """从反序列化的一个教室临时活动信息 json 中生成RoomTempActivity对象

        :param data: 反序列化成字典的教室临时活动 json
        :type data: Dict[str, Any]
        :return: 教室临时活动
        :rtype: RoomTempActivity
        """
        return RoomTempActivity(
            activity_info=RoomActivityInfo.from_dict(data),
            content=data['actContent'],
            department=data['actDepartment'],
            type=data['tempActType'],
            time_range=data['timeIn'],
            date=[date_from_str(date) for date in data['dateStr'].split(',')]
        )



from . import auth, course, exam, mycqu, score, card, exception, library, enroll
from .auth import *
from .course import *
from .exam import *
from .score import *
from .mycqu import *
from .card import *
from .library import *
from .enroll import *
from .exception import *
__all__ = ["auth", "course", "exam", "mycqu", "score", "card", "library", "enroll"]
__all__.extend(auth.__all__)
__all__.extend(course.__all__)
__all__.extend(exam.__all__)
__all__.extend(score.__all__)
__all__.extend(mycqu.__all__)
__all__.extend(card.__all__)
__all__.extend(library.__all__)
__all__.extend(enroll.__all__)
__all__.extend(exception.__all__)



from __future__ import annotations

import datetime
from typing import Any

from pydantic import BaseModel

from ...utils.datetimes import TIMEZONE


__all__ = ['Bill']


class Bill(BaseModel):
    """
    某次消费账单信息
    """
    name: str
    """交易名称"""
    date: datetime.datetime
    """交易时间"""
    place: str
    """交易地点"""
    tran_amount: float
    """交易金额"""
    acc_amount: float
    """账户余额"""

    @staticmethod
    def from_dict(data: dict[str, Any]) -> Bill:
        """
        从反序列化的（一个）账单 json 中获取账单信息

        :param data: json 反序列化得到的字典
        :type data: dict[str, Any]
        :return: 账单对象
        :rtype: Bill
        """
        return Bill(
            name=data['tranName'],
            date=datetime.datetime.strptime(
                data['tranDt'], "%Y-%m-%d %H:%M:%S").replace(tzinfo=TIMEZONE),
            place=data['mchAcctName'],
            tran_amount=float(data['tranAmt'] / 100),
            acc_amount=float(int(data['acctAmt']) / 100)
        )



from __future__ import annotations

from datetime import date
from typing import Optional, Any, List

from requests import Session
from pydantic import BaseModel

from .cqu_session import CQUSession
from ...exception import MycquUnauthorized
from ...utils.datetimes import date_from_str
from ...utils.request_transformer import Request, RequestTransformer

CUR_SESSION_URL = "https://my.cqu.edu.cn/api/resourceapi/session/cur-active-session"
ALL_SESSIONSINFO_URL = "https://my.cqu.edu.cn/api/resourceapi/session/list"
SESSION_DETAIL_URL = "https://my.cqu.edu.cn/api/resourceapi/session/detail/"


__all__ = ['CQUSessionInfo']

class CQUSessionInfo(BaseModel):
    """某学期的一些额外信息
    """
    session: CQUSession
    """对应的学期"""
    begin_date: Optional[date] = None
    """学期的开始日期"""
    end_date: Optional[date] = None
    """学期的结束日期"""

    @staticmethod
    def from_dict(data: dict[str, Any]) -> CQUSessionInfo:
        """从反序列化的（一个）学期信息 json 中获取学期信息

        :param data: json 反序列化得到的字典
        :type data: dict[str, Any]
        :return: 学期信息对象
        :rtype: CQUSessionInfo
        """
        res = CQUSessionInfo(
            session=CQUSession(year=data["year"],
                               is_autumn=data["term"] == "秋"),
            begin_date=date_from_str(data["beginDate"][0:10] if data["beginDate"] else None),
            end_date=date_from_str(data["endDate"][0:10] if data["endDate"] else None)
        )
        res.session.id = int(data["id"])

        return res

    @staticmethod
    @RequestTransformer.register()
    def _fetch_all(session: Request) -> List[CQUSessionInfo]:
        resp = yield session.get(ALL_SESSIONSINFO_URL)
        if resp.status_code == 401:
            raise MycquUnauthorized()
        cqusesions: List[CQUSessionInfo] = []
        for data in resp.json()['sessionVOList']:
            if not data['beginDate']:
                break
            cqusesions.append(CQUSessionInfo.from_dict(data))
        return cqusesions

    @staticmethod
    def fetch_all(session: Session) -> List[CQUSessionInfo]:
        """获取所有学期信息

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :return: 按时间降序排序的学期（最新学期可能尚未到来，其信息准确度也无法保障！）
        :rtype: List[CQUSessionInfo]
        """
        return CQUSessionInfo._fetch_all.sync_request(session)

    @staticmethod
    async def async_fetch_all(session: Request) -> List[CQUSessionInfo]:
        """
        异步的获取所有学期信息

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :return: 按时间降序排序的学期（最新学期可能尚未到来，其信息准确度也无法保障！）
        :rtype: List[CQUSessionInfo]
        """
        return await CQUSessionInfo._fetch_all.async_request(session)

    @staticmethod
    @RequestTransformer.register()
    def _fetch_detail(session: Request, cqu_session: CQUSession) -> CQUSessionInfo:
        if cqu_session.id is None:
            raise ValueError("cqu_session.id should not be None")
        resp = yield session.get(SESSION_DETAIL_URL + str(cqu_session.id))
        if resp.status_code == 401:
            raise MycquUnauthorized()
        return CQUSessionInfo.from_dict(resp.json()["data"])

    @staticmethod
    def fetch_detail(session: Session, cqu_session: CQUSession) -> CQUSessionInfo:
        """获取指定学期的详细信息

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :param cqu_session: 指定的学期
        :type cqu_session: CQUSession
        :return: 学期详细信息
        :rtype: CQUSessionInfo
        """
        return CQUSessionInfo._fetch_detail.sync_request(session, cqu_session)

    @staticmethod
    async def async_fetch_detail(session: Request, cqu_session: CQUSession) -> CQUSessionInfo:
        """
        异步的获取指定学期的详细信息

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :param cqu_session: 指定的学期
        :type cqu_session: CQUSession
        :return: 学期详细信息
        :rtype: CQUSessionInfo
        """
        return await CQUSessionInfo._fetch_detail.async_request(session, cqu_session)

    @staticmethod
    @RequestTransformer.register()
    def _fetch(session: Request) -> CQUSessionInfo:
        resp = yield session.get(CUR_SESSION_URL)
        if resp.status_code == 401:
            raise MycquUnauthorized()
        return CQUSessionInfo.from_dict(resp.json()["data"])

    @staticmethod
    def fetch(session: Session) -> CQUSessionInfo:
        """从 my.cqu.edu.cn 上获取当前学期的学期信息，需要登录并认证了 mycqu 的会话

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 认证
        :return: 本学期信息对象
        :rtype: CQUSessionInfo
        """
        return CQUSessionInfo._fetch.sync_request(session)

    @staticmethod
    async def async_fetch(session: Request) -> CQUSessionInfo:
        """
        异步的从 my.cqu.edu.cn 上获取当前学期的学期信息，需要登录并认证了 mycqu 的会话

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 认证
        :return: 本学期信息对象
        :rtype: CQUSessionInfo
        """
        return await CQUSessionInfo._fetch.async_request(session)



from __future__ import annotations

from typing import Any, Dict, Optional, Tuple, List, Union

from requests import Session

from pydantic import BaseModel

from .course import Course
from .course_day_time import CourseDayTime
from ..tools import get_course_raw, async_get_course_raw, get_enroll_raw, async_get_enroll_raw
from .cqu_session import CQUSession
from ...utils.datetimes import parse_weeks_str
from ...utils.period import Period
from ...utils.request_transformer import Request


__all__ = ['CourseTimetable']

class CourseTimetable(BaseModel):
    """课表对象，一个对象存储有相同课程、相同行课节次和相同星期的一批行课安排
    """
    course: Course
    """对应的课程"""
    stu_num: Optional[int] = None
    """学生数"""
    classroom: Optional[str] = None
    """行课地点，无则为 :obj:`None`"""
    weeks: List[Period]
    """行课周数，列表中每个元组 (a,b) 代表一个周数范围 a~b（包含 a, b），在单独的一周则有 b=a"""
    day_time: Optional[CourseDayTime] = None
    """行课的星期和节次，若时间是整周（如真实地占用整周的军训和某些实习、虚拟地使用一周的思修实践）
    则为 :obj:`None`"""
    whole_week: bool
    """是否真实地占用整周（如军训和某些实习是真实地占用、思修实践是“虚拟地占用”）"""
    classroom_name: Optional[str] = None
    """行课教室名称"""
    expr_projects: List[str]
    """实验课各次实验内容"""

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> CourseTimetable:
        """从反序列化的一个课表 json 中获取课表

        :param data: 反序列化成字典的课表 json
        :type data: Dict[str, Any]
        :return: 课表对象
        :rtype: CourseTimetable
        """
        return CourseTimetable(
            course=Course.from_dict(data),
            stu_num=data.get("selectedStuNum"),
            classroom=data.get("position"),
            weeks=parse_weeks_str(data.get("weeks")
                                  or data.get("teachingWeekFormat")),  # type: ignore
            day_time=CourseDayTime.from_dict(data),
            whole_week=bool(data["wholeWeekOccupy"]),
            classroom_name=data["roomName"],
            expr_projects=(data["exprProjectName"] or '').split(',')
        )

    @staticmethod
    def fetch(session: Session, code: str, cqu_session: Optional[Union[CQUSession, str]] = None) \
            -> List[CourseTimetable]:
        """从 my.cqu.edu.cn 上获取学生或老师的课表

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :param code: 学生或教师的学工号
        :type code: str
        :param cqu_session: 需要获取课表的学期，留空获取当前年级的课表
        :type cqu_session: Optional[Union[CQUSession, str]], optional
        :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
        :return: 获取的课表对象的列表
        :rtype: List[CourseTimetable]
        """
        resp = get_course_raw(session, code, cqu_session)
        return [CourseTimetable.from_dict(timetable) for timetable in resp
                if timetable["teachingWeekFormat"]
                ]

    @staticmethod
    async def async_fetch(session: Request, code: str, cqu_session: Optional[Union[CQUSession, str]] = None) \
            -> List[CourseTimetable]:
        """
        异步的从 my.cqu.edu.cn 上获取学生或老师的课表

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :param code: 学生或教师的学工号
        :type code: str
        :param cqu_session: 需要获取课表的学期，留空获取当前年级的课表
        :type cqu_session: Optional[Union[CQUSession, str]], optional
        :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
        :return: 获取的课表对象的列表
        :rtype: List[CourseTimetable]
        """
        resp = await async_get_course_raw(session, code, cqu_session)
        return [CourseTimetable.from_dict(timetable) for timetable in resp
                if timetable["teachingWeekFormat"]
                ]

    @staticmethod
    def fetch_enroll(session: Request) -> List[CourseTimetable]:
        """从 my.cqu.edu.cn 上获取学生已选课程

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
        :return: 获取的课表对象的列表
        :rtype: List[CourseTimetable]
        """
        res = get_enroll_raw(session)
        return [CourseTimetable.from_dict(timetable) for timetable in res]

    @staticmethod
    async def async_fetch_enroll(session: Request) -> List[CourseTimetable]:
        """
        异步的从 my.cqu.edu.cn 上获取学生已选课程

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
        :return: 获取的课表对象的列表
        :rtype: List[CourseTimetable]
        """
        res = await async_get_enroll_raw(session)
        return [CourseTimetable.from_dict(timetable) for timetable in res]



from __future__ import annotations

from typing import Dict, Any, Union, Optional, List

from requests import Session
from pydantic import BaseModel

from ..tools import get_score_raw, async_get_score_raw
from ...course import Course, CQUSession

__all__ = ['Score']


class Score(BaseModel):
    """
    成绩对象
    """
    session: CQUSession
    """学期"""
    course: Course
    """课程"""
    score: Optional[str] = None
    """成绩，可能为数字，也可能为字符（优、良等）"""
    study_nature: str
    """初修/重修"""
    course_nature: str
    """必修/选修"""

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> Score:
        """
        从反序列化的字典生成Score对象

        @param: data
        @type: dict
        @return: 返回成绩对象
        @rtype: Score
        """
        return Score(
            session=CQUSession.from_str(data["sessionName"]),
            course=Course.from_dict(data),
            score=data['effectiveScoreShow'],
            study_nature=data['studyNature'],
            course_nature=data['courseNature']
        )

    @staticmethod
    def fetch(auth: Union[str, Session], is_minor_boo: bool = False) -> List[Score]:
        """
        从网站获取成绩信息

        :param auth: 登陆后获取的 authorization 或者调用过 :func:`.mycqu.access_mycqu` 的 Session
        :type auth: Union[Session, str]
        :param is_minor_boo: 是否获取辅修成绩
        :type is_minor_boo: bool
        :return: 返回成绩对象
        :rtype: List[Score]
        :raises CQUWebsiteError: 查询时教务网报错
        """
        temp = get_score_raw(auth, is_minor_boo)
        score = []
        for courses in temp.values():
            for course in courses['stuScoreHomePgVoS']:
                score.append(Score.from_dict(course))
        return score

    @staticmethod
    async def async_fetch(auth: Union[str, Session], is_minor_boo: bool = False) -> List[Score]:
        """
        异步的从网站获取成绩信息

        :param auth: 登陆后获取的 authorization 或者调用过 :func:`.mycqu.access_mycqu` 的 Session
        :type auth: Union[Session, str]
        :param is_minor_boo: 是否获取辅修成绩
        :type is_minor_boo: bool
        :return: 返回成绩对象
        :rtype: List[Score]
        :raises CQUWebsiteError: 查询时教务网报错
        """
        temp = await async_get_score_raw(auth, is_minor_boo)
        score = []
        for courses in temp.values():
            for course in courses['stuScoreHomePgVoS']:
                score.append(Score.from_dict(course))
        return score


from __future__ import annotations

import re
from typing import List, Optional, ClassVar, Tuple

from pydantic import BaseModel

from ...course import CourseDayTime
from ...utils.datetimes import parse_weekday_str, parse_period_str, parse_weeks_str
from ...utils.period import Period


__all__ = ['EnrollCourseTimetable']

WEEKS_RE = re.compile("^(.*)周")
PERIOD_RE = re.compile("星期. [0-9]-[0-9]小节")
POS_RE = re.compile("&(.*)$")


class EnrollCourseTimetable(BaseModel):
    """
    可选具体课程上课时间、上课地点信息
    """
    weeks: List[Period]
    """上课周数"""
    time: Optional[CourseDayTime] = None
    """上课时间"""
    pos: Optional[str] = None
    """上课地点"""

    @staticmethod
    def from_str(data: str) -> List[EnrollCourseTimetable]:
        """从字符串中生成具体待选课程上课时间信息
        示例字符串"1-5,7-9周 星期二 6-7小节 &D1144 ;1-5,7-9周 星期五 3-4小节 &D1143 "

        :param data: 需提取信息的字符串
        :type data: str
        :return: 返回待选课程上课时间信息当列表
        :rtype: List[EnrollCourseTimetable]
        """
        items = data.split(';')
        result = []

        for item in items:
            pos_str = POS_RE.search(item)
            pos = None
            if pos_str:
                pos = pos_str.group().strip()[1:]

            period_str = PERIOD_RE.search(item)
            timetable = None
            if period_str:
                period_str = period_str.group()
                timetable = CourseDayTime(
                    weekday=parse_weekday_str(period_str[:3]),
                    period=parse_period_str(period_str[4:-2])
                )

            result.append(
                EnrollCourseTimetable(
                    weeks=parse_weeks_str(WEEKS_RE.search(item).group()[:-1]),
                    time=timetable,
                    pos=pos
                )
            )

        return result


from __future__ import annotations

import json
from typing import List, Dict

from requests import Session

from ..utils.request_transformer import Request, RequestTransformer

ENROLLMENT_COURSE_LIST_URL = "https://my.cqu.edu.cn/api/enrollment/enrollment/course-list?selectionSource="
ENROLLMENT_COURSE_DETAIL_URL = "https://my.cqu.edu.cn/api/enrollment/enrollment/courseDetails/"


__all__ = ['get_enroll_list_raw', 'get_enroll_detail_raw',
           'async_get_enroll_list_raw', 'async_get_enroll_detail_raw']


@RequestTransformer.register()
def _get_enroll_list_raw(session: Request, is_major: bool = True) -> Dict[str: List]:
    url = ENROLLMENT_COURSE_LIST_URL + ("主修" if is_major else "辅修")
    res = yield session.get(url)
    content = json.loads(res.text)
    assert content["status"] == "success"

    result = {}
    for data in content["data"]:
        result[data["selectionArea"]] = data["courseVOList"]

    return result

def get_enroll_list_raw(session: Session, is_major: bool = True) -> Dict[str: List]:
    """

    从 my.cqu.edu.cn 上获取学生可选课程

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
    :type session: Session
    :param is_major: 是否查询主修专业可选课程(为false时查询辅修专业可选课程)
    :type is_major: bool
    :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
    :return: 反序列化获取可选课程的的json
    :rtype: Dict[str, List]
    """

    return _get_enroll_list_raw.sync_request(session, is_major)

async def async_get_enroll_list_raw(session: Request, is_major: bool = True) -> Dict[str: List]:
    """

    异步的从 my.cqu.edu.cn 上获取学生可选课程

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
    :type session: Session
    :param is_major: 是否查询主修专业可选课程(为false时查询辅修专业可选课程)
    :type is_major: bool
    :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
    :return: 反序列化获取可选课程的的json
    :rtype: Dict[str, List]
    """

    return await _get_enroll_list_raw.async_request(session, is_major)

@RequestTransformer.register()
def _get_enroll_detail_raw(session: Request, course_id: str, is_major: bool = True) -> List:
    url = ENROLLMENT_COURSE_DETAIL_URL + course_id + "?selectionSource=" + ("主修" if is_major else "辅修")
    res = yield session.get(url)
    content = json.loads(res.text)
    return content['selectCourseListVOs'][0]['selectCourseVOList'] if len(content['selectCourseListVOs']) > 0 else []

def get_enroll_detail_raw(session: Session, course_id: str, is_major: bool = True) -> List:
    """
    从 my.cqu.edu.cn 上获取某可选课程详情

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
    :type session: Session
    :param course_id: 待查询课程参数
    :type course_id: str
    :param is_major: 待查询课程是否为主修课程(为false时表示查询辅修可选课程)
    :type is_major: bool
    :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
    :return: 反序列化查询到当可选课程的的json
    :rtype: List
    """
    return _get_enroll_detail_raw.sync_request(session, course_id, is_major)

async def async_get_enroll_detail_raw(session: Request, course_id: str, is_major: bool = True) -> List:
    """
    异步的从 my.cqu.edu.cn 上获取某可选课程详情

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
    :type session: Session
    :param course_id: 待查询课程参数
    :type course_id: str
    :param is_major: 待查询课程是否为主修课程(为false时表示查询辅修可选课程)
    :type is_major: bool
    :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
    :return: 反序列化查询到当可选课程的的json
    :rtype: List
    """
    return await _get_enroll_detail_raw.async_request(session, course_id, is_major)



import json
from html.parser import HTMLParser
from ..exception import TicketGetError, ParseError, CQUWebsiteError
from ..utils.request_transformer import Request, RequestTransformer


class _CardPageParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self._starttag: bool = False
        self.ssoticket_id: str = ""

    def handle_starttag(self, tag, attrs):
        if not self._starttag and tag == 'input' and ('name', 'ssoticketid') in attrs:
            self._starttag = True
            for key, val in attrs:
                if key == "value":
                    self.ssoticket_id = val
                    break


# 获取hallticket
@RequestTransformer.register()
def _get_hall_ticket(session: Request, ssoticket_id):
    url = 'http://card.cqu.edu.cn/cassyno/index'
    data = {
        'errorcode': '1',
        'continueurl': 'http://card.cqu.edu.cn/cassyno/index',
        'ssoticketid': ssoticket_id,
    }
    r = yield session.post(url, data=data)
    if r.status_code != 200:
        raise CQUWebsiteError()


# 利用登录之后的cookie获取一卡通的关键ticket
@RequestTransformer.register()
def _get_ticket(session: Request):
    url = 'http://card.cqu.edu.cn/Page/Page'
    data = {
        'EMenuName': '电费、网费',
        'MenuName': '电费、网费',
        'Url': 'http%3a%2f%2fcard.cqu.edu.cn%3a8080%2fblade-auth%2ftoken%2fthirdToToken%2ffwdt',
        'apptype': '4',
        'flowID': '10002'
    }
    r = yield session.post(url, data=data)
    if r.status_code != 200:
        raise CQUWebsiteError()
    ticket_start = r.text.find('ticket=')
    if ticket_start > 0:
        ticket_end = r.text.find("'", ticket_start)
        ticket = r.text[ticket_start + len('ticket='): ticket_end]
        return ticket
    else:
        raise TicketGetError()


# 利用ticket获取一卡通关键cookie
@RequestTransformer.register()
def _get_synjones_auth(session: Request, ticket: str):
    url = 'http://card.cqu.edu.cn:8080/blade-auth/token/fwdt'
    data = {'ticket': ticket, 'json': 'true'}
    r = yield session.post(url, data=data)
    if r.status_code != 200:
        raise CQUWebsiteError()
    try:
        dic = json.loads(r.text)
        token = dic['data']['access_token']
    except:
        raise ParseError()
    else:
        return 'bearer ' + token


# 利用关键cookie获取水电费dic
@RequestTransformer.register()
def _get_fee_data(session: Request, synjones_auth, room, fee_item_id):
    url = "http://card.cqu.edu.cn:8080/charge/feeitem/getThirdData"
    data = {
        'feeitemid': fee_item_id,
        'json': 'true',
        'level': '2',
        'room': room,
        'type': 'IEC',
    }
    cookie = {'synjones-auth': synjones_auth}
    r = yield session.post(url, data=data, cookies=cookie)
    if r.status_code != 200:
        raise CQUWebsiteError()
    dic = json.loads(r.text)
    if dic['msg'] == 'success':
        return dic
    else:
        raise CQUWebsiteError(dic['msg'])



from __future__ import annotations
from typing import Optional, Union

from requests import Session
from .models.cqu_session import CQUSession
from .models.cqu_session_info import CQUSessionInfo
from ..exception import MycquUnauthorized
from ..utils.request_transformer import Request, RequestTransformer

TIMETABLE_URL = "https://my.cqu.edu.cn/api/timetable/class/timetable/student/my-table-detail"

__all__ = ['get_course_raw', 'async_get_course_raw', 'get_enroll_raw', 'async_get_enroll_raw']


@RequestTransformer.register()
def _get_course_raw(session: Request, code: str, cqu_session: Optional[Union[CQUSession, str]] = None):
    if cqu_session is None:
        cqu_session = (yield CQUSessionInfo._fetch).session
    elif isinstance(cqu_session, str):
        cqu_session = CQUSession.from_str(cqu_session)
    assert isinstance(cqu_session, CQUSession)
    resp = yield session.post(TIMETABLE_URL,
                              params={"sessionId": (yield cqu_session._get_id)},
                              json=[code],
                              )
    if resp.status_code == 401:
        raise MycquUnauthorized()
    result = resp.json().get('classTimetableVOList')
    return result if result is not None else []

@RequestTransformer.register()
def _get_enroll_raw(session: Request):
    res = yield session.get(f'https://my.cqu.edu.cn/api/enrollment/timetable/student')
    result = res.json().get('data')
    return result if result is not None else []

def get_course_raw(session: Session, code: str, cqu_session: Optional[Union[CQUSession, str]] = None):
    """从 my.cqu.edu.cn 上获取学生或老师的课表

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
    :type session: Session
    :param code: 学生或教师的学工号
    :type code: str
    :param cqu_session: 需要获取课表的学期，留空获取当前年级的课表
    :type cqu_session: Optional[Union[CQUSession, str]], optional
    :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
    :return: 反序列化获取课表的json
    :rtype: dict
    """
    return _get_course_raw.sync_request(session, code, cqu_session)

async def async_get_course_raw(session: Request, code: str, cqu_session: Optional[Union[CQUSession, str]] = None):
    """
    异步的从 my.cqu.edu.cn 上获取学生或老师的课表

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
    :type session: Session
    :param code: 学生或教师的学工号
    :type code: str
    :param cqu_session: 需要获取课表的学期，留空获取当前年级的课表
    :type cqu_session: Optional[Union[CQUSession, str]], optional
    :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
    :return: 反序列化获取课表的json
    :rtype: dict
    """
    return await _get_course_raw.async_request(session, code, cqu_session)

def get_enroll_raw(session: Request):
    """
    从 my.cqu.edu.cn 上获取学生的选课信息

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
    :type session: Session
    :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
    :return: 反序列化获取课表的json
    :rtype: dict
    """
    return _get_enroll_raw.sync_request(session)

async def async_get_enroll_raw(session: Request):
    """
    异步的从 my.cqu.edu.cn 上获取学生的选课信息

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
    :type session: Session
    :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
    :return: 反序列化获取课表的json
    :rtype: dict
    """
    return await _get_enroll_raw.async_request(session)



from enum import Enum
from typing import NewType, Tuple, Optional, Dict, Protocol, Any, Literal, TypeVar

from .params_mapper import RequestParamsMapper


__all__ = ['ResponseProtocol', 'RequestProtocol', 'RequestReturns', 'RequestParams', 'Requestable', 'Request', 'Response']

REQUEST_METHOD = Literal['delete', 'get', 'head', 'options', 'patch', 'post', 'put']


class ResponseProtocol(Protocol):
    """
    具有`__await__`、`ok`、`status_code`、`url`、`content`、`text`、`json`属性的对象
    """
    def __await__(self): ...

    @property
    def ok(self) -> bool: ...

    @property
    def status_code(self) -> int: ...

    @property
    def url(self) -> str: ...

    @property
    def content(self) -> bytes: ...

    @property
    def text(self) -> str: ...

    @property
    def json(self, **kwargs: Any) -> Any: ...

    @property
    def headers(self) -> Any: ...

class RequestProtocol(Protocol):
    """
    具有网络请求相关方法，所有请求方法均返回一个满足`Response`协议的对象
    """
    def request(*args, **kwargs) -> ResponseProtocol: ...

    def get(*args, **kwargs) -> ResponseProtocol: ...

    def post(*args, **kwargs) -> ResponseProtocol: ...

    def put(*args, **kwargs) -> ResponseProtocol: ...

    def patch(*args, **kwargs) -> ResponseProtocol: ...

    def delete(*args, **kwargs) -> ResponseProtocol: ...
    def options(*args, **kwargs) -> ResponseProtocol: ...

    def head(*args, **kwargs) -> ResponseProtocol: ...

Request = TypeVar('Request', bound=RequestProtocol)
Response = TypeVar('Response', bound=ResponseProtocol)

class RequestParams:
    """
    调用了`RequestTransformer`中register方法的函数均应通过yield返回此类以支持对不同请求库的参数映射
    """
    def __init__(self, *, method: REQUEST_METHOD, url: str, allow_redirects: bool = True, **kwargs):
        """
        :param method: 请求方式，'delete', 'get', 'head', 'options', 'patch', 'post', 'put'中的某一个
        :param url: 请求url
        :param allow_redirects: 运行自动重定向
        """
        self.method = method
        self.url = url
        self.allow_redirects = allow_redirects
        self.other_params = kwargs

    def to_param_dict(self, param_mapper: RequestParamsMapper) -> Dict:
        """
        通过`RequestParamsMapper`类将请求参数映射为不同请求库需求的参数字典
        """
        result = {}
        result.update({param_mapper.method.value: self.method})
        result.update({param_mapper.url.value: self.url})
        result.update({param_mapper.allow_redirects.value: self.allow_redirects})
        for k, v in self.other_params.items():
            param_key: Optional[Enum] = getattr(param_mapper, k, None)
            if param_key is None:
                raise Exception('请求参数未在RequestsParamsMapper中给出')
            result.update({param_key.value: v})
        return result

RequestReturns = NewType('RequestReturns', Tuple[RequestProtocol, RequestParams])

class Requestable:
    def __init__(self, requestable: RequestProtocol):
        self.requestable = requestable
    def request(self, method: REQUEST_METHOD, url: str, *, allow_redirects: bool = True, **kwargs) -> RequestReturns:
        return RequestReturns((self.requestable, RequestParams(url=url, method=method, allow_redirects=allow_redirects,
                                                               **kwargs)))

    def get(self, url: str, *, allow_redirects: bool = True, **kwargs) -> RequestReturns:
        return self.request('get', url, allow_redirects=allow_redirects, **kwargs)

    def post(self, url: str, *, allow_redirects: bool = True, **kwargs) -> RequestReturns:
        return self.request('post', url, allow_redirects=allow_redirects, **kwargs)

    def put(self, url: str, *, allow_redirects: bool = True, **kwargs) -> RequestReturns:
        return self.request('put', url, allow_redirects=allow_redirects, **kwargs)

    def patch(self, url: str, *, allow_redirects: bool = True, **kwargs) -> RequestReturns:
        return self.request('patch', url, allow_redirects=allow_redirects, **kwargs)

    def delete(self, url: str, *, allow_redirects: bool = True, **kwargs) -> RequestReturns:
        return self.request('delete', url, allow_redirects=allow_redirects, **kwargs)

    def options(self, url: str, *, allow_redirects: bool = True, **kwargs) -> RequestReturns:
        return self.request('options', url, allow_redirects=allow_redirects, **kwargs)

    def head(self, url: str, *, allow_redirects: bool = True, **kwargs) -> RequestReturns:
        return self.request('head', url, allow_redirects=allow_redirects, **kwargs)


from .enroll_course_info import EnrollCourseInfo
from .enroll_course_item import EnrollCourseItem
from .enroll_course_timetable import EnrollCourseTimetable

__all__ = ['EnrollCourseInfo', 'EnrollCourseItem', 'EnrollCourseTimetable']



from __future__ import annotations

from typing import Dict, Any, Union, Optional

from requests import Session
from pydantic import BaseModel

from ..tools import get_gpa_ranking_raw, async_get_gpa_ranking_raw

__all__ = ['GpaRanking']


class GpaRanking(BaseModel):
    """
    绩点对象
    """
    gpa: float
    """学生总绩点"""
    major_ranking: Optional[int] = None
    """专业排名"""
    grade_ranking: Optional[int] = None
    """年级排名"""
    class_ranking: Optional[int] = None
    """班级排名"""
    weighted_avg: float
    """加权平均分"""
    minor_weighted_avg: Optional[float] = None
    """辅修加权平均分"""
    minor_gpa: Optional[float] = None
    """辅修绩点"""

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> GpaRanking:
        """
        从反序列化的字典生成GpaRanking对象

        @param: data
        @type: dict
        @return: 返回绩点排名对象
        @rtype: GpaRanking
        """
        return GpaRanking(
            gpa=float(data['gpa']),
            major_ranking=data['majorRanking'] and int(data['majorRanking']),
            grade_ranking=data['gradeRanking'] and int(data['gradeRanking']),
            class_ranking=data['classRanking'] and int(data['classRanking']),
            weighted_avg=float(data['weightedAvg']),
            minor_weighted_avg=data['minorWeightedAvg'] and float(data['minorWeightedAvg']),
            minor_gpa=data['minorGpa'] and float(data['minorGpa']),
        )

    @staticmethod
    def fetch(auth: Union[str, Session]) -> GpaRanking:
        """
        从网站获取绩点排名信息

        :param auth: 登陆后获取的 authorization 或者调用过 :func:`.mycqu.access_mycqu` 的 Session
        :type auth: Union[Session, str]
        :return: 返回绩点排名对象
        :rtype: GpaRanking
        :raises CQUWebsiteError: 查询时教务网报错
        """
        return GpaRanking.from_dict(get_gpa_ranking_raw(auth))

    @staticmethod
    async def async_fetch(auth: Union[str, Session]) -> GpaRanking:
        """
        异步的从网站获取绩点排名信息

        :param auth: 登陆后获取的 authorization 或者调用过 :func:`.mycqu.access_mycqu` 的 Session
        :type auth: Union[Session, str]
        :return: 返回绩点排名对象
        :rtype: GpaRanking
        :raises CQUWebsiteError: 查询时教务网报错
        """
        return GpaRanking.from_dict(await async_get_gpa_ranking_raw(auth))



"""
成绩相关模块
"""
from __future__ import annotations

import json
from typing import Dict, Union, Optional, Generic

import requests

from ..exception import CQUWebsiteError, MycquUnauthorized
from ..utils.request_transformer import Request, RequestTransformer

__all__ = ("get_score_raw", "async_get_score_raw", "get_gpa_ranking_raw", "async_get_gpa_ranking_raw")


def _launch_authorized_header(authorization: str) -> Dict:
    return {
        'Referer': 'https://my.cqu.edu.cn/sam/home',
        'User-Agent': 'Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.0; Trident/4.0)',
        'Authorization': authorization
    }

@RequestTransformer.register()
def _get_score_raw(request: Request, is_minor_boo: bool, headers: Optional[Dict] = None):
    url = 'https://my.cqu.edu.cn/api/sam/score/student/score' + ('?isMinorBoo=true' if is_minor_boo else '')
    res = yield request.get(url, headers=headers)

    content = json.loads(res.content)
    if res.status_code == 401:
        raise MycquUnauthorized()
    if content['status'] == 'error':
        raise CQUWebsiteError(content['msg'])
    return content['data']


def get_score_raw(auth: Union[Generic[Request], str], is_minor_boo: bool = False) -> Dict:
    """
    获取学生原始成绩

    :param auth: 登陆后获取的authorization或者调用过mycqu.access_mycqu的session
    :type auth: Union[Session, str]
    :param is_minor_boo: 是否获取辅修成绩
    :type is_minor_boo: bool
    :return: 反序列化获取的score列表
    :rtype: Dict
    """
    if not isinstance(auth, str):
        return _get_score_raw.sync_request(auth, is_minor_boo)
    else:
        return _get_score_raw.sync_request(requests, is_minor_boo, _launch_authorized_header(auth))

async def async_get_score_raw(session: Generic[Request], is_minor_boo: bool = False):
    """
    异步的获取学生原始成绩

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
    :type session: Session
    :param is_minor_boo: 是否获取辅修成绩
    :type is_minor_boo: bool
    :return: 反序列化获取的score列表
    :rtype: Dict
    """
    return await _get_score_raw.async_request(session, is_minor_boo)

@RequestTransformer.register()
def _get_gpa_ranking_raw(request: Request, headers: Optional[Dict] = None):
    res = yield request.get('https://my.cqu.edu.cn/api/sam/score/student/studentGpaRanking', headers=headers)

    content = json.loads(res.content)
    if res.status_code == 401:
        raise MycquUnauthorized()
    if content['status'] == 'error':
        raise CQUWebsiteError(content['msg'])
    return content['data']


def get_gpa_ranking_raw(auth: Union[Generic[Request], str]):
    """
    获取学生绩点排名

    :param auth: 登陆后获取的authorization或者调用过mycqu.access_mycqu的session
    :type auth: Union[Session, str]
    :return: 反序列化获取的绩点、排名
    :rtype: Dict
    """
    if not isinstance(auth, str):
        return _get_gpa_ranking_raw.sync_request(auth)
    else:
        headers = _launch_authorized_header(auth)
        return _get_gpa_ranking_raw.sync_request(headers)

async def async_get_gpa_ranking_raw(session: Generic[Request]):
    """
    异步的获取学生绩点排名

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
    :type session: Union[Session, str]
    :return: 反序列化获取的绩点、排名
    :rtype: Dict
    """
    return await _get_gpa_ranking_raw.async_request(session)



from typing import List, Tuple, Dict, Optional
from datetime import date, time, datetime
import pytz

from ..utils.period import Period

TIMEZONE = datetime.now(pytz.timezone("Asia/Shanghai")).tzinfo
SHORT_WEEKDAY: Dict[str, int] = {
    "一": 0,
    "二": 1,
    "三": 2,
    "四": 3,
    "五": 4,
    "六": 5,
    "日": 6
}
LONG_WEEKDAY: Dict[str, int] = {
    "星期一": 0,
    "星期二": 1,
    "星期三": 2,
    "星期四": 3,
    "星期五": 4,
    "星期六": 5,
    "星期日": 6
}


def time_from_str(string: str) -> time:
    hour, minute = map(int, string.split(":"))
    return time(hour, minute, tzinfo=TIMEZONE)


def parse_period_str(string: str) -> Period:
    period = tuple(map(int, string.split("-")))
    assert len(period) == 1 or len(period) == 2
    return Period(start=period[0], end=(period[1] if len(period) == 2 else period[0]))


def parse_weeks_str(string: str) -> List[Period]:
    return [parse_period_str(unit) for unit in string.split(',')]


def parse_weekday_str(string: str) -> Optional[int]:
    return SHORT_WEEKDAY.get(string) if SHORT_WEEKDAY.get(string) is not None else LONG_WEEKDAY.get(string)


def date_from_str(string: str) -> Optional[date]:
    return date.fromisoformat(string) if string is not None else None


def datetime_from_str(string: str) -> datetime:
    return datetime.strptime(string, '%Y-%m-%d %H:%M:%S').replace(tzinfo=TIMEZONE)



from __future__ import annotations

import re
from typing import ClassVar, Tuple, List, Optional

import requests
from requests import Session
from pydantic import BaseModel

from ...utils.request_transformer import Request, RequestTransformer
from ...exception import CQUSessionIdNotExist

CQUSESSIONS_URL = "https://my.cqu.edu.cn/api/timetable/optionFinder/session?blankOption=false"
SESSION_RE = re.compile("^([0-9]{4})年?(春|秋)$")


__all__ = ['CQUSession']


class CQUSession(BaseModel):
    """重大的某一学期
    """
    id: Optional[int] = None
    """学期ID"""
    year: int
    """主要行课年份"""
    is_autumn: bool
    """是否为秋冬季学期"""

    def __str__(self):
        return str(self.year) + ('秋' if self.is_autumn else '春')

    @RequestTransformer.register()
    def _get_id(self, client: Request) -> int:
        if self.id is not None:
            return self.id
        else:
            sessions: List[CQUSession] = yield self._fetch
            res = list(filter(lambda x: x.year == self.year and x.is_autumn == self.is_autumn, sessions))
            if len(res) == 0:
                raise CQUSessionIdNotExist
            else:
                return res[0].id

    def get_id(self, client: Request) -> int:
        return self._get_id.sync_request(client)

    async def async_get_id(self, client: Request) -> int:
        return self._get_id.async_request(client)

    @staticmethod
    def from_str(string: str, id: Optional[int] = None) -> CQUSession:
        """
        从学期字符串中解析学期，如果直接调用该方法，在获取id时会自动进行一次网络请求

        >>> CQUSession.from_str("2021春")
        CQUSession(year=2021, is_autumn=False)
        >>> CQUSession.from_str("2020年秋")
        CQUSession(year=2020, is_autumn=True)

        :param string: 学期字符串，如“2021春”、“2020年秋”
        :type string: str
        :param id: 学期id，应交由CQUSession方法自动设置
        :type id: Optional[int]
        :raises ValueError: 字符串不是一个预期中的学期字符串时抛出
        :return: 对应的学期
        :rtype: CQUSession
        """
        match = SESSION_RE.match(string)
        if match:
            result = CQUSession(
                year=match[1],
                is_autumn=match[2] == "秋"
            )
            result.id = id
            return result
        else:
            raise ValueError(f"string {string} is not a session")

    @staticmethod
    @RequestTransformer.register()
    def _fetch(request: Request) -> List[CQUSession]:
        session_list = []
        for session in (yield request.get(CQUSESSIONS_URL)).json():
            session_list.append(CQUSession.from_str(session["name"], int(session["id"])))
        return session_list

    @staticmethod
    def fetch(session: Optional[Session] = None) -> List[CQUSession]:
        """从 my.cqu.edu.cn 上获取各个学期

        :return: 各个学期组成的列表
        :rtype: List[CQUSession]
        """
        return CQUSession._fetch.sync_request(requests if session is None else session)

    @staticmethod
    async def async_fetch(session: Request) -> List[CQUSession]:
        """
        异步的从 my.cqu.edu.cn 上获取各个学期

        :return: 各个学期组成的列表
        :rtype: List[CQUSession]
        """
        return await CQUSession._fetch.async_request(session)



"""
统一身份认证相关的模块
"""

from typing import Optional, Callable, Generic
from requests import Session, Response

from ._authserver import *
from ._sso import *
from ..exception import NeedCaptcha, InvaildCaptcha
from ..utils.request_transformer import Request


__all__ = ['is_logined', 'logout', 'access_service', 'login',
           'async_is_logined', 'async_logout', 'async_access_service', 'async_login']


def is_logined(session: Session, use_sso: bool = True) -> bool:
    """判断是否处于统一身份认证登陆状态

    :param session: 会话
    :type session: Session
    :param use_sso: 是否使用 sso 而非 authserver, 默认为 :obj::`True`
    :type use_sso: bool, optional
    :return: :obj:`True` 如果处于登陆状态，:obj:`False` 如果处于未登陆或登陆过期状态
    :rtype: bool
    """
    return is_sso_logined(session) if use_sso else is_authserver_logined(session)

async def async_is_logined(session: Generic[Request], use_sso: bool = True) -> bool:
    """
    异步的判断是否处于统一身份认证登陆状态

    :param session: 会话
    :type session: Session
    :param use_sso: 是否使用 sso 而非 authserver, 默认为 :obj::`True`
    :type use_sso: bool, optional
    :return: :obj:`True` 如果处于登陆状态，:obj:`False` 如果处于未登陆或登陆过期状态
    :rtype: bool
    """
    return await async_is_sso_logined(session) if use_sso else await async_is_authserver_logined(session)


def logout(session: Session, use_sso: bool = True) -> None:
    """注销统一身份认证登录状态

    :param session: 进行过登录的会话
    :type session: Session
    :param use_sso: 是否使用 sso 而非 authserver, 默认为 :obj::`True`
    :type use_sso: bool, optional
    """
    logout_sso(session) if use_sso else logout_authserver(session)

async def async_logout(session: Generic[Request], use_sso: bool = True) -> None:
    """
    异步的注销统一身份认证登录状态

    :param session: 进行过登录的会话
    :type session: Session
    :param use_sso: 是否使用 sso 而非 authserver, 默认为 :obj::`True`
    :type use_sso: bool, optional
    """
    await async_logout_sso(session) if use_sso else await async_logout_authserver(session)


def access_service(session: Session, service: str, use_sso: bool = True) -> Response:
    """从登录了统一身份认证的会话获取指定服务的许可

    :param session: 登录了统一身份认证的会话
    :type session: Session
    :param service: 服务的 url
    :type service: str
    :param use_sso: 是否使用 sso 而非 authserver, 默认为 :obj::`True`
    :type use_sso: bool, optional
    :raises NotLogined: 统一身份认证未登录时抛出
    :return: 访问服务 url 的 :class:`Response`
    :rtype: Response
    """
    return access_sso_service(session, service) if use_sso else access_authserver_service(session, service)

async def async_access_service(session: Generic[Request], service: str, use_sso: bool = True) -> Response:
    """
    异步的从登录了统一身份认证的会话获取指定服务的许可

    :param session: 登录了统一身份认证的会话
    :type session: Session
    :param service: 服务的 url
    :type service: str
    :param use_sso: 是否使用 sso 而非 authserver, 默认为 :obj::`True`
    :type use_sso: bool, optional
    :raises NotLogined: 统一身份认证未登录时抛出
    :return: 访问服务 url 的 :class:`Response`
    :rtype: Response
    """
    return await async_access_sso_service(session, service) if use_sso else await async_access_authserver_service(session, service)


def login(session: Session,
          username: str,
          password: str,
          service: Optional[str] = None,
          timeout: int = 10,
          force_relogin: bool = False,
          captcha_callback: Optional[
              Callable[[bytes, str], Optional[str]]] = None,
          keep_longer: bool = False,
          kick_others: bool = False,
          use_sso: bool = True
          ) -> Response:
    """登录统一身份认证

    :param session: 用于登录统一身份认证的会话
    :type session: Session
    :param username: 统一身份认证号或学工号
    :type username: str
    :param password: 统一身份认证密码
    :type password: str
    :param service: 需要登录的服务，默认（:obj:`None`）则先不登陆任何服务
    :type service: Optional[str], optional
    :param timeout: 连接超时时限，默认为 10（单位秒）
    :type timeout: int, optional
    :param force_relogin: 强制重登，当会话中已经有有效的登陆 cookies 时依然重新登录，默认为 :obj:`False`
    :type force_relogin: bool, optional
    :param captcha_callback: 需要输入验证码时调用的回调函数，默认为 :obj:`None` 即不设置回调；
                             当需要输入验证码，但回调没有设置或回调返回 :obj:`None` 时，抛出异常 :class:`NeedCaptcha`；
                             该函数接受一个 :class:`bytes` 型参数为验证码图片的文件数据，一个 :class:`str` 型参数为图片的 MIME 类型，
                             返回验证码文本或 :obj:`None`。
    :type captcha_callback: Optional[Callable[[bytes, str], Optional[str]]], optional
    :param keep_longer: 保持更长时间的登录状态（保持一周）
    :type keep_longer: bool
    :param kick_others: 当目标用户开启了“单处登录”并有其他登录会话时，踢出其他会话并登录单前会话；若该参数为 :obj:`False` 则抛出
                       :class:`MultiSessionConflict`
    :type kick_others: bool
    :param use_sso: 是否使用 sso 而非 authserver, 默认为 :obj::`True`
    :type use_sso: bool, optional
    :raises UnknownAuthserverException: 未知认证错误
    :raises InvaildCaptcha: 无效的验证码
    :raises IncorrectLoginCredentials: 错误的登陆凭据（如错误的密码、用户名）
    :raises NeedCaptcha: 需要提供验证码，获得验证码文本之后可调用所抛出异常的 :func:`NeedCaptcha.after_captcha` 函数来继续登陆
    :raises MultiSessionConflict: 和其他会话冲突
    :return: 登陆了统一身份认证后所跳转到的地址的 :class:`Response`
    :rtype: Response
    """
    try:
        return login_sso(session, username, password, service, timeout, force_relogin) \
            if use_sso else login_authserver(session, username, password, service, timeout, force_relogin, keep_longer,
                                             kick_others)
    except NeedCaptcha as e:
        if captcha_callback is None:
            raise e
        else:
            captcha_str = captcha_callback(e.image, e.image_type)
            if captcha_str is None:
                raise InvaildCaptcha()
            else:
                return e.after_captcha(captcha_str)

async def async_login(session: Generic[Request],
          username: str,
          password: str,
          service: Optional[str] = None,
          timeout: int = 10,
          force_relogin: bool = False,
          captcha_callback: Optional[
              Callable[[bytes, str], Optional[str]]] = None,
          keep_longer: bool = False,
          kick_others: bool = False,
          use_sso: bool = True
          ) -> Response:
    """
    异步的登录统一身份认证

    :param session: 用于登录统一身份认证的会话
    :type session: Session
    :param username: 统一身份认证号或学工号
    :type username: str
    :param password: 统一身份认证密码
    :type password: str
    :param service: 需要登录的服务，默认（:obj:`None`）则先不登陆任何服务
    :type service: Optional[str], optional
    :param timeout: 连接超时时限，默认为 10（单位秒）
    :type timeout: int, optional
    :param force_relogin: 强制重登，当会话中已经有有效的登陆 cookies 时依然重新登录，默认为 :obj:`False`
    :type force_relogin: bool, optional
    :param captcha_callback: 需要输入验证码时调用的回调函数，默认为 :obj:`None` 即不设置回调；
                             当需要输入验证码，但回调没有设置或回调返回 :obj:`None` 时，抛出异常 :class:`NeedCaptcha`；
                             该函数接受一个 :class:`bytes` 型参数为验证码图片的文件数据，一个 :class:`str` 型参数为图片的 MIME 类型，
                             返回验证码文本或 :obj:`None`。
    :type captcha_callback: Optional[Callable[[bytes, str], Optional[str]]], optional
    :param keep_longer: 保持更长时间的登录状态（保持一周）
    :type keep_longer: bool
    :param kick_others: 当目标用户开启了“单处登录”并有其他登录会话时，踢出其他会话并登录单前会话；若该参数为 :obj:`False` 则抛出
                       :class:`MultiSessionConflict`
    :type kick_others: bool
    :param use_sso: 是否使用 sso 而非 authserver, 默认为 :obj::`True`
    :type use_sso: bool, optional
    :raises UnknownAuthserverException: 未知认证错误
    :raises InvaildCaptcha: 无效的验证码
    :raises IncorrectLoginCredentials: 错误的登陆凭据（如错误的密码、用户名）
    :raises NeedCaptcha: 需要提供验证码，获得验证码文本之后可调用所抛出异常的 :func:`NeedCaptcha.after_captcha` 函数来继续登陆
    :raises MultiSessionConflict: 和其他会话冲突
    :return: 登陆了统一身份认证后所跳转到的地址的 :class:`Response`
    :rtype: Response
    """
    try:
        return await async_login_sso(session, username, password, service, timeout, force_relogin) \
            if use_sso else await async_login_authserver(session, username, password, service, timeout, force_relogin, keep_longer,
                                             kick_others)
    except NeedCaptcha as e:
        if captcha_callback is None:
            raise e
        else:
            captcha_str = captcha_callback(e.image, e.image_type)
            if captcha_str is None:
                raise InvaildCaptcha()
            else:
                return e.after_captcha(captcha_str)



from __future__ import annotations
from typing import Any, Dict, Optional, Union

from pydantic import BaseModel

from .cqu_session import CQUSession


__all__ = ['Course']

class Course(BaseModel):
    """与具体行课时间无关的课程信息
    """
    name: Optional[str] = None
    """课程名称"""
    code: Optional[str] = None
    """课程代码"""
    course_num: Optional[str] = None
    """教学班号，在无法获取时（如考表 :class:`.exam.Exam` 中）设为 :obj:`None`"""
    dept: Optional[str] = None
    """开课学院， 在无法获取时（如成绩 :class:`.score.Score`中）设为 :obj:`None`"""
    credit: Optional[float] = None
    """学分，无法获取到则为 :obj:`None`（如在考表 :class:`.exam.Exam` 中）"""
    instructor: Optional[str] = None
    """教师"""
    session: Optional[CQUSession] = None
    """学期，无法获取时则为 :obj:`None`"""

    @staticmethod
    def from_dict(data: Dict[str, Any],
                  session: Optional[Union[str, CQUSession]] = None) -> Course:
        """从反序列化的（一个）课表或考表 json 中返回课程

        :param data: 反序列化成字典的课表或考表 json
        :type data: Dict[str, Any]
        :param session: 学期字符串或学期对象，留空则尝试从 ``data`` 中获取
        :type session: Optional[Union[str, CQUSession]], optional
        :return: 对应的课程对象
        :rtype: Course
        """
        if session is None and not data.get("session") is None:
            session = CQUSession.from_str(data["session"])
        if isinstance(session, str):
            session = CQUSession.from_str(session)
        assert isinstance(session, CQUSession) or session is None

        instructor_name = None
        if data.get("instructorName") is not None:
            instructor_name = data.get("instructorName")
        elif data.get("instructorNames") is not None:
            instructor_name = data.get("instructorNames")
        elif data.get('classTimetableInstrVOList') is not None:
            instructor_name = ', '.join(instructor.get('instructorName')
                                        for instructor in data.get('classTimetableInstrVOList'))

        return Course(
            name=data["courseName"],
            code=data["courseCode"],
            course_num=data.get("classNbr"),
            dept=data.get(
                "courseDepartmentName") or data.get("courseDeptShortName"),
            credit=data.get("credit") or data.get("courseCredit"),
            instructor=instructor_name,
            session=session,
        )



from base64 import b64encode, b64decode
from typing import Optional, Dict, Generic

from requests import Session

from ..utils.request_transformer import RequestTransformer, Request, Response
from ..exception import IncorrectLoginCredentials, InvaildCaptcha, UnknownAuthserverException
from .._lib_wrapper.encrypt import des_ecb_encryptor, pad8
from ._page_parser import _SSOPageParser, _SSOErrorParser
from ._authorizer import Authorizer

_SSO_CAPTCHA_ERROR_CODE = 1320007
_SSO_ERROR_CODES = {1030027: '用户名或密码错误，请确认后重新输入',
                    1030031: '用户名或密码错误，请确认后重新输入',
                    1410041: '当前用户名已失效',
                    1410040: '当前用户名已失效',
                    1320007: '验证码有误，请确认后重新输入'}

__all__ = ['is_sso_logined', 'logout_sso', 'access_sso_service', 'login_sso',
           'async_is_sso_logined', 'async_logout_sso', 'async_access_sso_service', 'async_login_sso']

def is_sso_logined(session: Session) -> bool:
    """判断是否处于统一身份认证（sso）登陆状态

    :param session: 会话
    :type session: Session
    :return: :obj:`True` 如果处于登陆状态，:obj:`False` 如果处于未登陆或登陆过期状态
    :rtype: bool
    """
    return SSOAuthorizer[Session].is_logined(session)

async def async_is_sso_logined(session: Request) -> bool:
    """
    异步的判断是否处于统一身份认证（sso）登陆状态

    :param session: 会话
    :type session: Session
    :return: :obj:`True` 如果处于登陆状态，:obj:`False` 如果处于未登陆或登陆过期状态
    :rtype: bool
    """
    return await SSOAuthorizer.async_is_logined(session)

def logout_sso(session: Session) -> None:
    """注销统一身份认证（sso）登录状态

    :param session: 进行过登录的会话
    :type session: Session
    """
    return SSOAuthorizer[Session].logout(session)

async def async_logout_sso(session: Request) -> None:
    """
    异步的注销统一身份认证（sso）登录状态

    :param session: 进行过登录的会话
    :type session: Session
    """
    return await SSOAuthorizer.async_logout(session)

def access_sso_service(session: Session, service: str) -> Response:
    """从登录了统一身份认证（sso）的会话获取指定服务的许可

    :param session: 登录了统一身份认证的会话
    :type session: Session
    :param service: 服务的 url
    :type service: str
    :raises NotLogined: 统一身份认证未登录时抛出
    :return: 访问服务 url 的 :class:`Response`
    :rtype: Response
    """
    return SSOAuthorizer[Session].access_service(session, service)

async def async_access_sso_service(session: Request, service: str) -> Response:
    """
    异步的从登录了统一身份认证（sso）的会话获取指定服务的许可

    :param session: 登录了统一身份认证的会话
    :type session: Session
    :param service: 服务的 url
    :type service: str
    :raises NotLogined: 统一身份认证未登录时抛出
    :return: 访问服务 url 的 :class:`Response`
    :rtype: Response
    """
    return await SSOAuthorizer.async_access_service(session, service)


class SSOAuthorizer(Authorizer, Generic[Request]):
    def __init__(self,
                 session: Request,
                 username: str,
                 password: str,
                 service: Optional[str] = None,
                 timeout: int = 10,
                 force_relogin: bool = False,
                 keep_longer: bool = False,
                 kick_others: bool = False):

        super().__init__(session, username, password, service, timeout, force_relogin, keep_longer, kick_others)

        self._login_res = None
        self._page_data = None

    LOGIN_URL = "https://sso.cqu.edu.cn/login"
    LOGOUT_URL = "https://sso.cqu.edu.cn/logout"
    ROOT_URL = "https://sso.cqu.edu.cn"

    @RequestTransformer.register()
    def _get_request_data(self, session: Request) -> Dict:
        resp = yield session.get(
            self.LOGIN_URL,
            params=self.service and {"service": self.service},
            allow_redirects=False,
            timeout=self.timeout
        )
        if resp.status_code == 302:
            if self.force_relogin:
                yield self._logout
                resp = yield session.get(self.LOGIN_URL, timeout=self.timeout)
            else:
                self._login_res = yield session.get(resp.headers['Location'], allow_redirects=False, timeout=self.timeout)
                return {}
        if resp.status_code != 200:
            UnknownAuthserverException(
                f"status code {resp.status_code} is got (302 expected) when sending login post, "
                "but can not find the element span.login_auth_error#msg")

        page_data = _SSOPageParser().parse(resp.text)
        croypto = page_data['login-croypto']
        passwd_encrypted = b64encode(des_ecb_encryptor(b64decode(croypto))(pad8(self.password.encode())))
        passwd_encrypted = str(passwd_encrypted, encoding='utf-8')
        request_data = {
            'username': self.username,
            'type': 'UsernamePassword',
            '_eventId': 'submit',
            'geolocation': '',
            'execution': page_data['login-page-flowkey'],
            'croypto': croypto,
            'password': passwd_encrypted
        }

        return request_data

    @RequestTransformer.register()
    def _need_captcha(self, session: Request) -> Optional[str]:
        if self._page_data is not None:
            return f"{self.ROOT_URL}/{self._page_data['captcha-url']}"

    def _need_captcha_handler(self, captcha: str, request_data: Dict):
        request_data['captcha_code'] = [captcha, captcha]

    @RequestTransformer.register()
    def _login(self, session: Request, request_data: Dict) -> Response:
        if self._login_res is not None:
            return (yield session.get(self._login_res.headers['Location'], allow_redirects=False, timeout=self.timeout))

        login_resp = yield session.post(self.LOGIN_URL,
                                  params=self.service and {"service": self.service},
                                  data=request_data,
                                  allow_redirects=False,
                                  timeout=self.timeout)
        if login_resp.status_code == 302:
            return (yield session.get(login_resp.headers['Location'], allow_redirects=False, timeout=self.timeout))
        elif login_resp.status_code == 401:
            raise IncorrectLoginCredentials()
        elif login_resp.status_code == 200:
            error_code: Optional[int] = _SSOErrorParser().parse(login_resp.text)
            if error_code == _SSO_CAPTCHA_ERROR_CODE:
                raise InvaildCaptcha()
            elif error_code is None:
                raise UnknownAuthserverException("No error code")
            else:
                raise UnknownAuthserverException(
                    f"{error_code}: {_SSO_ERROR_CODES.get(error_code, '')}")


def login_sso(session: Session,
              username: str,
              password: str,
              service: Optional[str] = None,
              timeout: int = 10,
              force_relogin: bool = False
              ):
    """登录统一身份认证（sso）

    :param session: 用于登录统一身份认证的会话
    :type session: Session
    :param username: 统一身份认证号或学工号
    :type username: str
    :param password: 统一身份认证密码
    :type password: str
    :param service: 需要登录的服务，默认（:obj:`None`）则先不登陆任何服务
    :type service: Optional[str], optional
    :param timeout: 连接超时时限，默认为 10（单位秒）
    :type timeout: int, optional
    :param force_relogin: 强制重登，当会话中已经有有效的登陆 cookies 时依然重新登录，默认为 :obj:`False`
    :type force_relogin: bool, optional
    :raises InvaildCaptcha: 无效的验证码
    :raises IncorrectLoginCredentials: 错误的登陆凭据（如错误的密码、用户名）
    :raises NeedCaptcha: 需要提供验证码，获得验证码文本之后可调用所抛出异常的 :func:`NeedCaptcha.after_captcha` 函数来继续登陆
    :return: 登陆了统一身份认证后所跳转到的地址的 :class:`Response`
    :rtype: Response
    """
    return SSOAuthorizer[Session].login(session, username, password, service, timeout, force_relogin)

async def async_login_sso(session: Request,
              username: str,
              password: str,
              service: Optional[str] = None,
              timeout: int = 10,
              force_relogin: bool = False
              ):
    """
    异步的登录统一身份认证（sso）

    :param session: 用于登录统一身份认证的会话
    :type session: Session
    :param username: 统一身份认证号或学工号
    :type username: str
    :param password: 统一身份认证密码
    :type password: str
    :param service: 需要登录的服务，默认（:obj:`None`）则先不登陆任何服务
    :type service: Optional[str], optional
    :param timeout: 连接超时时限，默认为 10（单位秒）
    :type timeout: int, optional
    :param force_relogin: 强制重登，当会话中已经有有效的登陆 cookies 时依然重新登录，默认为 :obj:`False`
    :type force_relogin: bool, optional
    :raises InvaildCaptcha: 无效的验证码
    :raises IncorrectLoginCredentials: 错误的登陆凭据（如错误的密码、用户名）
    :raises NeedCaptcha: 需要提供验证码，获得验证码文本之后可调用所抛出异常的 :func:`NeedCaptcha.after_captcha` 函数来继续登陆
    :return: 登陆了统一身份认证后所跳转到的地址的 :class:`Response`
    :rtype: Response
    """
    return await SSOAuthorizer.async_login(session, username, password, service, timeout, force_relogin)



from __future__ import annotations

from typing import List, Dict, Optional

from pydantic import BaseModel

from .enroll_course_timetable import EnrollCourseTimetable
from ..tools import get_enroll_detail_raw, async_get_enroll_detail_raw
from ...course import Course

from requests import Session


__all__ = ['EnrollCourseItem']

class EnrollCourseItem(BaseModel):
    """
    可选具体课程，包含课程上课时间、上课教师、教室可容纳学生等信息
    """
    id: Optional[str] = None
    """可选具体课程id，每个可选具体课程具有唯一id，部分从属课程该值为`None`"""
    session_id: Optional[str] = None
    """可选具体课程所在学期ID，部分从属课程该值为`None`"""
    checked: Optional[bool] = None
    """是否已经选择该课程，部分从属课程该值为`None`"""
    course_id: Optional[str] = None
    """该具体课程所属课程ID，部分从属课程该值为`None`"""
    course: Course
    """课程信息"""
    type: str
    """具体课程类别，如：理论、实验"""
    selected_num: Optional[int] = None
    """已选课程学生，部分从属课程该值为`None`"""
    capacity: Optional[int] = None
    """该课程容量上限，部分从属课程该值为`None`"""
    children: Optional[List[EnrollCourseItem]] = None
    """该课程从属课程列表，一般为理论课程的实验课"""
    campus: Optional[str] = None
    """所属校区，如D区，部分从属课程该值为`None`"""
    parent_id: Optional[str] = None
    """所从属具体课程id，如果不存在从属关系，该值为None"""
    timetables: List[EnrollCourseTimetable]

    @staticmethod
    def from_dict(data: Dict) -> EnrollCourseItem:
        """
        从反序列化的可选具体课程字典中返回`EnrollCourseItem`
        :param data: 反序列化成字典的可选具体课程
        :type data: Dict[str, Any]
        :return: 对应的可选具体课程列表
        :rtype: EnrollCourseItem
        """
        return EnrollCourseItem(
            id=data['id'],
            session_id=data['sessionId'],
            checked=data['checked'],
            course_id=data['courseId'],
            course=Course.from_dict(data),
            type=data['classType'],
            selected_num=data['selectedNum'],
            capacity=data['stuCapacity'],
            children=[EnrollCourseItem.from_dict(item) for item in data['childrenList']]
                     if data['childrenList'] else None,
            campus=data['campusShortName'],
            parent_id=data['parentClassId'],
            timetables=EnrollCourseTimetable.from_str(data['classTime']) if data['classTime'] is not None else []
        )

    @staticmethod
    def fetch(session: Session, id: str, is_major: bool = True) -> List[EnrollCourseItem]:
        """从 my.cqu.edu.cn 上获取学生可选具体课程

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :param id: 需要获取的课程id（非Course Code）
        :type id: str
        :param is_major: 是否获取主修可选课程，为`False`时查询辅修可选课程
        :type is_major: bool
        :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
        :return: 获取的可选具体课程对象的列表
        :rtype: List[EnrollCourseItem]
        """
        res = get_enroll_detail_raw(session, id, is_major)
        return [EnrollCourseItem.from_dict(item) for item in res]

    @staticmethod
    async def async_fetch(session: Session, id: str, is_major: bool = True) -> List[EnrollCourseItem]:
        """
        异步的从 my.cqu.edu.cn 上获取学生可选具体课程

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :param id: 需要获取的课程id（非Course Code）
        :type id: str
        :param is_major: 是否获取主修可选课程，为`False`时查询辅修可选课程
        :type is_major: bool
        :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
        :return: 获取的可选具体课程对象的列表
        :rtype: List[EnrollCourseItem]
        """
        res = await async_get_enroll_detail_raw(session, id, is_major)
        return [EnrollCourseItem.from_dict(item) for item in res]



from abc import ABC, abstractmethod
from typing import Optional, Callable, Dict, NoReturn, Union, Awaitable, Generic
from functools import partial

from ..exception import NeedCaptcha, InvaildCaptcha, NotLogined
from ..utils.request_transformer import Request, Response, RequestTransformer


class Authorizer(ABC, Generic[Request]):
    """
    SSO 和 Authorizer 登陆验证流程均为：获取登陆信息->判断是否需要填写验证码
    ->需要填写，抛出:class:`NeedCaptcha`异常
    ->无需填写，依靠登陆信息进行登陆
    该类要求子类实现获取登陆信息、判断是否需要填写验证码、获取验证码后应当如何修改登陆信息、使用登陆信息进行登陆四个函数，
    组合上述四个方法实现了通用的login、logout、is_logined、access_service方法
    """
    def __init__(
            self,
            session: Request,
            username: str,
            password: str,
            service: Optional[str] = None,
            timeout: int = 10,
            force_relogin: bool = False,
            keep_longer: bool = False,
            kick_others: bool = False
    ):
        self.session = session
        self.username = username
        self.password = password
        self.service = service
        self.timeout = timeout
        self.force_relogin = force_relogin
        self.keep_longer = keep_longer
        self.kick_others = kick_others

    LOGIN_URL = ''
    LOGOUT_URL = ''

    @abstractmethod
    def _get_request_data(self, session: Request) -> Dict:
        """
        获取请求所需的相关参数
        """
        pass

    @abstractmethod
    def _need_captcha(self, session: Request) -> Optional[str]:
        """
        是否需要验证码，如果需要，则返回验证码目标url
        """
        pass

    @abstractmethod
    def _need_captcha_handler(self, captcha: str, request_data: Dict):
        """
        拥有验证码之后应该如何修改请求参数
        """
        pass

    @abstractmethod
    def _login(self, session: Request, request_data: Dict) -> Response:
        """
        通过获取的请求参数进行登陆
        """
        pass

    def _raise_need_captcha(
            self, url: str,
            after_captcha: Callable[[str], Union[Response, Awaitable[Response]]], timeout: int = 10) -> NoReturn:
        """
        抛出`NeedCaptcha`异常
        """
        captcha_img = self.session.get(url, timeout=timeout)
        raise NeedCaptcha(captcha_img.content,
                          captcha_img.headers["Content-Type"],
                          after_captcha)


    @classmethod
    def login(cls,
              session: Request,
              username: str,
              password: str,
              service: Optional[str] = None,
              timeout: int = 10,
              force_relogin: bool = False,
              keep_longer: bool = False,
              kick_others: bool = False
              ) -> Response:
        """
        组合登陆流程
        """
        authorizer = cls(session, username, password, service, timeout, force_relogin, keep_longer, kick_others)
        request_data = authorizer._get_request_data.sync_request(authorizer.session)

        is_need_captcha = authorizer._need_captcha.sync_request(authorizer.session)
        if is_need_captcha is not None:
            request_data_changer = partial(authorizer._need_captcha_handler, request_data=request_data)
            after_captcha = lambda captcha: authorizer._login.sync_request(authorizer.session, request_data_changer(captcha))
            authorizer._raise_need_captcha(is_need_captcha, after_captcha, authorizer.timeout)

        return authorizer._login.sync_request(authorizer.session, request_data)

    @classmethod
    async def async_login(
            cls,
            session: Request,
            username: str,
            password: str,
            service: Optional[str] = None,
            timeout: int = 10,
            force_relogin: bool = False,
            keep_longer: bool = False,
            kick_others: bool = False
    ) -> Response:
        """
        组合登陆流程
        """
        authorizer = cls(session, username, password, service, timeout, force_relogin, keep_longer, kick_others)
        request_data = await authorizer._get_request_data.async_request(authorizer.session)

        is_need_captcha = await authorizer._need_captcha.async_request(authorizer.session)
        if is_need_captcha is not None:
            request_data_changer = partial(authorizer._need_captcha_handler, request_data=request_data)
            after_captcha = lambda captcha: authorizer._login.sync_request(authorizer.session, request_data_changer(captcha))
            authorizer._raise_need_captcha(is_need_captcha, after_captcha, authorizer.timeout)

        return await authorizer._login.async_request(authorizer.session, request_data)

    @classmethod
    @RequestTransformer.register()
    def _is_logined(cls, session: Request) -> bool:
        assert cls.LOGIN_URL != '', '子类未重写`IS_LOGINED_URL`'
        res = yield session.get(cls.LOGIN_URL, allow_redirects=False)
        return res.status_code == 302

    @classmethod
    def is_logined(cls, session: Request) -> bool:
        """
        判断是否处于统一身份认证登陆状态

        :param session: 会话
        :type session: Session
        :return: :obj:`True` 如果处于登陆状态，:obj:`False` 如果处于未登陆或登陆过期状态
        :rtype: bool
        """
        return cls._is_logined.sync_request(session)

    @classmethod
    async def async_is_logined(cls, session: Request) -> bool:
        """
        异步的判断是否处于统一身份认证登陆状态

        :param session: 会话
        :type session: Session
        :return: :obj:`True` 如果处于登陆状态，:obj:`False` 如果处于未登陆或登陆过期状态
        :rtype: bool
        """
        return await cls._is_logined.async_request(session)


    @classmethod
    @RequestTransformer.register()
    def _access_service(cls, session: Request, service: str) -> Response:
        assert cls.LOGIN_URL != '', '子类未重写`ACCESS_SERVICE_URL`'
        resp = yield session.get(cls.LOGIN_URL, params={"service": service}, allow_redirects=False)
        if resp.status_code != 302:
            # TODO
            raise NotLogined()
        res = yield session.get(url=resp.headers['Location'], allow_redirects=False)
        return res

    @classmethod
    def access_service(cls, session: Request, service: str) -> Response:
        """
        从登录了统一身份认证的会话获取指定服务的许可

        :param session: 登录了统一身份认证的会话
        :type session: Session
        :param service: 服务的 url
        :type service: str
        :raises NotLogined: 统一身份认证未登录时抛出
        :return: 访问服务 url 的 :class:`Response`
        :rtype: Response
        """
        return cls._access_service.sync_request(session, service)

    @classmethod
    async def async_access_service(cls, session: Request, service: str) -> Response:
        """
        异步的从登录了统一身份认证的会话获取指定服务的许可

        :param session: 登录了统一身份认证的会话
        :type session: Session
        :param service: 服务的 url
        :type service: str
        :raises NotLogined: 统一身份认证未登录时抛出
        :return: 访问服务 url 的 :class:`Response`
        :rtype: Response
        """
        return await cls._access_service.async_request(session, service)

    @classmethod
    @RequestTransformer.register()
    def _logout(cls, session: Request) -> None:
        assert cls.LOGOUT_URL != '', '子类未重写`LOGOUT_URL`'
        yield session.get(cls.LOGOUT_URL)

    @classmethod
    def logout(cls, session: Request) -> None:
        """
        注销统一身份认证（sso）登录状态

        :param session: 进行过登录的会话
        :type session: Session
        """
        cls._logout.sync_request(session)

    @classmethod
    async def async_logout(cls, session: Request) -> None:
        """
        异步的注销统一身份认证（sso）登录状态

        :param session: 进行过登录的会话
        :type session: Session
        """
        await cls._logout.async_request(session)



from __future__ import annotations

from typing import Optional, Any

from requests import Session
from pydantic import BaseModel

from ..tools import get_fees_raw, async_get_fees_raw
from ...utils.request_transformer import Request


__all__ = ['EnergyFees']


class EnergyFees(BaseModel):
    """
    某宿舍的水电费相关信息
    """
    balance: float
    """账户余额"""
    electricity_subsidy: Optional[float] = None
    """电剩余补助（仅虎溪校区拥有）"""
    water_subsidy: Optional[float] = None
    """水剩余补助（仅虎溪校区拥有）"""
    subsidies: Optional[float] = None
    """补助余额（仅老校区拥有）"""

    @staticmethod
    def from_dict(data: dict[str, Any], is_huxi: bool) -> EnergyFees:
        """从反序列化的（一个）水电费 json 中获取水电费信息

        :param data: json 反序列化得到的字典
        :type data: dict[str, Any]
        :param is_huxi: 目标寝室是否在虎溪校区
        :type is_huxi: bool
        :return: 学期信息对象
        :rtype: EnergyFees
        """
        if is_huxi:
            return EnergyFees(
                balance=data["剩余金额"],
                electricity_subsidy=data["电剩余补助"],
                water_subsidy=data["水剩余补助"],
                subsidies=None,
            )
        else:
            return EnergyFees(
                balance=data["现金余额"],
                electricity_subsidy=None,
                water_subsidy=None,
                subsidies=data["补贴余额"],
            )

    @staticmethod
    def fetch(session: Session, is_huxi: bool, room: str) -> EnergyFees:
        """从 card.cqu.edu.cn 上获取当前水电费信息，需要登录了统一身份认证的会话

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 card.cqu.edu.cn 进行了认证（:func:`.card.access_card`）的 requests 会话
        :type session: Session
        :param is_huxi: 房间号是否为虎溪校区的房间
        :type is_huxi: bool
        :param room: 需要获取水电费详情的宿舍
        :type room: str
        :raises NetworkError: 当访问相关网页时statue code不为200时抛出
        :raises TicketGetError: 当未能从网页对应位置中获取到ticket时抛出
        :raises ParseError: 当从返回数据解析所需值失败时抛出
        :raises CQUWebsiteError: 当网页获取水电费状态码不为success时抛出
        :return: 返回相关宿舍的水电费信息
        :rtype: EnergyFees
        """
        return EnergyFees.from_dict(get_fees_raw(session, is_huxi, room)["map"]["showData"], is_huxi)

    @staticmethod
    async def async_fetch(session: Request, is_huxi: bool, room: str) -> EnergyFees:
        """从 card.cqu.edu.cn 上获取当前水电费信息，需要登录了统一身份认证的会话

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 card.cqu.edu.cn 进行了认证（:func:`.card.access_card`）的 requests 会话
        :type session: Session
        :param is_huxi: 房间号是否为虎溪校区的房间
        :type is_huxi: bool
        :param room: 需要获取水电费详情的宿舍
        :type room: str
        :raises NetworkError: 当访问相关网页时statue code不为200时抛出
        :raises TicketGetError: 当未能从网页对应位置中获取到ticket时抛出
        :raises ParseError: 当从返回数据解析所需值失败时抛出
        :raises CQUWebsiteError: 当网页获取水电费状态码不为success时抛出
        :return: 返回相关宿舍的水电费信息
        :rtype: EnergyFees
        """
        return EnergyFees.from_dict((await async_get_fees_raw(session, is_huxi, room))["map"]["showData"], is_huxi)



from typing import Dict, Any

from pydantic import BaseModel

from .room_activity_info import RoomActivityInfo


class RoomCourse(BaseModel):
    """教室课程活动"""
    activity_info: RoomActivityInfo
    """基础信息"""
    class_number: str
    """教学班号"""
    course_code: str
    """课程代码"""
    course_name: str
    """课程名称"""
    department: str
    """开课学院"""
    stu_num: int
    """选课学生数"""
    credit: float
    """课程学分"""
    instructor_name: str
    """教师姓名"""

    @staticmethod
    def from_dict(data: Dict[str, Any]):
        """从反序列化的一个教室课程信息 json 中生成RoomCourse对象

        :param data: 反序列化成字典的教室课程 json
        :type data: Dict[str, Any]
        :return: 教室课程
        :rtype: RoomCourse
        """
        return RoomCourse(
            activity_info=RoomActivityInfo.from_dict(data),
            class_number=data['classNbr'],
            course_code=data['courseCode'],
            course_name=data['courseName'],
            department=data['courseDepartmentName'],
            stu_num=int(data['selectedStuNum']),
            credit=float(data['credit']),
            instructor_name=data['instructorName'],
        )


from __future__ import annotations

from typing import Dict, Any, Optional, List
from datetime import date, time

from pydantic import BaseModel

from .invigilator import Invigilator
from ..tools import get_exam_raw, async_get_exam_raw
from ...course import Course
from ...utils.datetimes import date_from_str, time_from_str
from ...utils.request_transformer import Request


__all__ = ['Exam']


class Exam(BaseModel):
    """考试信息
    """
    course: Course
    """考试对应的课程，其中学分 :attr:`credit`、教师 :attr:`instructor`、教学班号 :attr:`course_num` 可能无法获取（其值会设置为 :obj:`None`）"""
    batch: str
    """考试批次，如 :obj:`"非集中考试周"`"""
    batch_id: int
    """选课系统中考试批次的内部id"""
    building: str
    """考场楼栋"""
    floor: Optional[int] = None
    """考场楼层"""
    room: str
    """考场地点"""
    stu_num: int
    """考场人数"""
    date: date
    """考试日期"""
    start_time: time
    """考试开始时间"""
    end_time: time
    """考试结束时间"""
    week: int
    """周次"""
    weekday: int
    """星期，0为周一，6为周日"""
    stu_id: str
    """考生学号"""
    seat_num: int
    """考生座号"""
    chief_invi: List[Invigilator]
    """监考员"""
    asst_invi: Optional[List[Invigilator]] = None
    """副监考员"""

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> Exam:
        """从反序列化后的 json 数据中的一次考试数据生成 :class:`Exam` 对象

        :param data: 反序列化后的 json 数据中的一次考试数据
        :type data: Dict[str, Any]
        :return: 对应的 :class:`Exam` 对象
        :rtype: [type]
        """
        course = Course.from_dict(data)
        return Exam(
            course=course,
            batch=data["batchName"],
            batch_id=data["batchId"],
            building=data["buildingName"],
            room=data["roomName"],
            floor=data["floorNum"] if data["floorNum"] != "null" else None,
            date=date_from_str(data["examDate"]),
            start_time=time_from_str(data["startTime"]),
            end_time=time_from_str(data["endTime"]),
            week=data["week"],
            weekday=int(data["weekDay"]) - 1,
            stu_id=data["studentId"],
            seat_num=data["seatNum"],
            stu_num=data["examStuNum"],
            chief_invi=[Invigilator.from_dict(invi)
                        for invi in data["simpleChiefinvigilatorVOS"]]
            if data['simpleChiefinvigilatorVOS'] is not None else [],
            asst_invi=data["simpleAssistantInviVOS"] and ([Invigilator.from_dict(invi)
                                                           for invi in data["simpleAssistantInviVOS"]]
                                                          if data['simpleAssistantInviVOS'] is not None else [])
        )

    @staticmethod
    def fetch(session: Request, student_id: str) -> List[Exam]:
        """从 my.cqu.edu.cn 上获取指定学生的考表

        :param session: 登陆后获取的 authorization 或者调用过 :func:`.mycqu.access_mycqu` 的 Session
        :type session: Request
        :param student_id: 学生学号
        :type student_id: str
        :return: 本学期的考表
        :rtype: List[Exam]
        """
        return [Exam.from_dict(exam)
                for exam in get_exam_raw(student_id, session)["data"]]

    @staticmethod
    async def async_fetch(session: Request, student_id: str) -> List[Exam]:
        """从 my.cqu.edu.cn 上获取指定学生的考表

        :param session: 登陆后获取的 authorization 或者调用过 :func:`.mycqu.access_mycqu` 的 Session
        :type session: Request
        :param student_id: 学生学号
        :type student_id: str
        :return: 本学期的考表
        :rtype: List[Exam]
        """
        return [Exam.from_dict(exam)
                for exam in (await async_get_exam_raw(session, student_id))["data"]]



"""从两种模块不同的模块名字中加载加密模块
"""
from typing import Callable
try:
    from Cryptodome.Cipher import AES as _AES
    from Cryptodome.Cipher import DES as _DES
    from Cryptodome.Util.Padding import pad as _pad
except ImportError:
    try:
        from Crypto.Cipher import AES as _AES  # type: ignore
        from Crypto.Cipher import DES as _DES  # type: ignore
        from Crypto.Util.Padding import pad as _pad  # type: ignore
    except ImportError:
        raise ImportError(  # pylint: ignore disable=raise-missing-from
            "Please install pyryptodome, pyryptodomex")


def pad16(x: bytes) -> bytes:
    return _pad(x, 16, style='pkcs7')


def pad8(x: bytes) -> bytes:
    return _pad(x, 8, style='pkcs7')


def aes_cbc_encryptor(key: bytes, iv: bytes) -> Callable[[bytes], bytes]:
    return _AES.new(key=key, iv=iv, mode=_AES.MODE_CBC).encrypt


def aes_ecb_encryptor(key: bytes) -> Callable[[bytes], bytes]:
    return _AES.new(key, _AES.MODE_ECB).encrypt


def des_ecb_encryptor(key: bytes) -> Callable[[bytes], bytes]:
    return _DES.new(key, _DES.MODE_ECB).encrypt


__all__ = ("aes_cbc_encryptor", "aes_ecb_encryptor",
           "des_ecb_encryptor", "pad16", "pad8")



from .room import Room
from .room_activity_info import RoomActivityInfo
from .room_course import RoomCourse
from .room_exam import RoomExam
from .room_exam_invigilator import RoomExamInvigilator
from .room_temp_activity import RoomTempActivity
from .room_timetable import RoomTimetable

__all__ = ['Room', 'RoomActivityInfo', 'RoomCourse', 'RoomExam', 'RoomExamInvigilator', 'RoomTempActivity', 'RoomTimetable']



from __future__ import annotations
from typing import Any, Dict, Optional, Tuple

from pydantic import BaseModel

from ...utils.datetimes import parse_period_str, parse_weekday_str
from ...utils.period import Period


__all__ = ['CourseDayTime']

class CourseDayTime(BaseModel):
    """课程一次的星期和节次
    """
    weekday: int
    """星期，0 为周一，6 为周日，此与 :attr:`datetime.date.day` 一致"""
    period: Period
    """节次，第一个元素为开始节次，第二个元素为结束节次（该节次也包括在范围内）。
    只有一节课时，两个元素相同。
    """

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> Optional[CourseDayTime]:
        """从反序列化的（一个）课表 json 中获取课程的星期和节次

        :param data: 反序列化成字典的课表 json
        :type data: Dict[str, Any]
        :return: 若其中有课程的星期和节次则返回相应对象，否则返回 :obj:`None`
        :rtype: Optional[CourseDayTime]
        """
        if data.get("periodFormat") and data.get("weekDayFormat"):
            return CourseDayTime(
                weekday=parse_weekday_str(data["weekDayFormat"]),
                period=parse_period_str(data["periodFormat"])
            )
        return None



from mycqu.auth import login
from requests import Session, Response

USERNAME = '07102028'
PASSWORD = 'Zhud626300'

def access_sso_service(session: Session, service: str):
    class ArriveService(Exception):
        def __init__(self, url: str):
            super().__init__()
            self.url: str = url

    def response_hook(response: Response, *args, **kwargs):
        #print(response.status_code, response.url, response.headers.get('Location'))
        if response.url.startswith("https://sso.cqu.edu.cn/login") and response.status_code != 302:
            assert '<div class="code">adapter</div>' in response.text
            response.headers['Location'] = './clientredirect?client_name=adapter'
            response.status_code = 302
            return response
        if response.url.startswith("https://sso.cqu.edu.cn/login") and response.status_code == 302 and response.headers['Location'].startswith(service):
            raise ArriveService(response.headers['Location'])

    try:
        assert not session.get("https://sso.cqu.edu.cn/login",
                               params={"service": service},
                               allow_redirects=True,
                               hooks={"response": response_hook}
                               )
    except ArriveService as arrive_service:
        return arrive_service.url

session = Session()
login(session, USERNAME, PASSWORD)
print(access_sso_service(session, "https://atrust.cqu.edu.cn:443/passport/v1/auth/cas"))



from typing import Dict, Any, List

from pydantic import BaseModel

from .room_activity_info import RoomActivityInfo
from .room_exam_invigilator import RoomExamInvigilator

__all__ = ['RoomExam']


class RoomExam(BaseModel):
    """教室中考试活动相关信息"""
    activity_info: RoomActivityInfo
    """基础信息"""
    course_name: str
    """考试课程名称"""
    stu_capacity: int
    """考试学生数量"""
    time_range: str
    """持续时间"""
    invigilators: List[RoomExamInvigilator]
    """监考员信息"""

    @staticmethod
    def from_dict(data: Dict[str, Any]):
        """从反序列化的一个教室考试信息 json 中生成RoomExam对象

        :param data: 反序列化成字典的教室考试 json
        :type data: Dict[str, Any]
        :return: 教室考试
        :rtype: RoomExam
        """
        return RoomExam(
            activity_info=RoomActivityInfo.from_dict(data),
            course_name=data['courseName'],
            stu_capacity=int(data['stuCapacity']),
            time_range=data['timeIn'],
            invigilators=[RoomExamInvigilator.from_dict(temp) for temp in data['invigilatorVOList']]
        )



from __future__ import annotations

from typing import List

from requests import Session
from pydantic import BaseModel

from .bill import Bill
from ..tools import get_card_raw, get_bill_raw, async_get_card_raw, async_get_bill_raw
from ...utils.request_transformer import Request


__all__ = ['Card']


class Card(BaseModel):
    """
    校园卡及其账单信息
    """
    id: str
    """校园卡id"""
    amount: float
    """账户余额"""

    @staticmethod
    def fetch(session: Session) -> Card:
        """
        从card.cqu.edu.cn获取校园卡信息

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 card.cqu.edu.cn 进行了认证（:func:`.card.access_card`）的 requests 会话
        :type session: Session
        :raises CQUWebsiteError: 当网页获取状态码不为0000时抛出
        :return: 获取的校园卡信息
        :rtype: Card
        """
        card_info = get_card_raw(session)

        return Card(
            id=card_info['acctNo'],
            amount=float(card_info['acctAmt'] / 100)
        )

    @staticmethod
    async def async_fetch(session: Request) -> Card:
        """
        异步的从card.cqu.edu.cn获取校园卡信息

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 card.cqu.edu.cn 进行了认证（:func:`.card.access_card`）的 requests 会话
        :type session: Session
        :raises CQUWebsiteError: 当网页获取状态码不为0000时抛出
        :return: 获取的校园卡信息
        :rtype: Card
        """
        card_info = await async_get_card_raw(session)

        return Card(
            id=card_info['acctNo'],
            amount=float(card_info['acctAmt'] / 100)
        )

    def fetch_bills(self, session: Session) -> List[Bill]:
        """
        从card.cqu.edu.cn获取校园卡账单

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 card.cqu.edu.cn 进行了认证（:func:`.card.access_card`）的 requests 会话
        :type session: Session
        :return: 获取的校园卡账单信息
        :rtype: dict
        """
        bill_info = get_bill_raw(session, self.id, 30)
        bills = []
        for bill in bill_info:
            bills.append(Bill.from_dict(bill))

        return bills

    async def async_fetch_bills(self, session: Request) -> List[Bill]:
        """
        异步的从card.cqu.edu.cn获取校园卡账单

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 card.cqu.edu.cn 进行了认证（:func:`.card.access_card`）的 requests 会话
        :type session: Session
        :return: 获取的校园卡账单信息
        :rtype: dict
        """
        bill_info = await async_get_bill_raw(session, self.id, 30)
        bills = []
        for bill in bill_info:
            bills.append(Bill.from_dict(bill))

        return bills


import random
import re
from base64 import b64encode
from html.parser import HTMLParser
from typing import Dict, Optional

from .._lib_wrapper.encrypt import pad16, aes_cbc_encryptor
from mycqu.exception import NotAllowedService, InvaildCaptcha, IncorrectLoginCredentials, \
    UnknownAuthserverException, ParseError

_CHAR_SET = 'ABCDEFGHJKMNPQRSTWXYZabcdefhijkmnprstwxyz2345678'


class _AuthPageParser(HTMLParser):
    _SALT_RE: re.Pattern = re.compile('var pwdDefaultEncryptSalt = "([^"]+)"')

    def __init__(self):
        super().__init__()
        self.input_data: Dict[str, Optional[str]] = \
            {'lt': None, 'dllt': None,
                'execution': None, '_eventId': None, 'rmShown': None}
        """几个关键的标签数据"""
        self.salt: Optional[str] = None
        """加密所用的盐"""
        self._js_start: bool = False
        self._js_end: bool = False
        self._error: bool = False
        self._error_head: bool = False

    def handle_starttag(self, tag, attrs):
        if tag == 'input':
            name: Optional[str] = None
            value: Optional[str] = None
            for attr in attrs:
                if attr[0] == 'name':
                    if attr[1] in self.input_data:
                        name = attr[1]
                    else:
                        break
                elif attr[0] == 'value':
                    value = attr[1]
            if name:
                self.input_data[name] = value
        elif tag == 'script' and attrs and attrs[0] == ("type", "text/javascript"):
            self._js_start = True
        elif tag == "div" and attrs == [("id", "msg"), ("class", "errors")]:
            self._error = True
        elif tag == 'h2' and self._error:
            self._error_head = True

    def handle_data(self, data):
        if self._js_start and not self._js_end:
            match = self._SALT_RE.search(data)
            if match:
                self.salt = match[1]
            self._js_end = True
        elif self._error_head:
            error_str = data.strip()
            if error_str == "应用未注册":
                raise NotAllowedService(error_str)
            raise UnknownAuthserverException(
                "Error message before login: "+error_str)


class _SSOPageParser(HTMLParser):
    _SALT_RE: re.Pattern = re.compile('var pwdDefaultEncryptSalt = "([^"]+)"')

    class _AllValuesGot(Exception):
        pass

    def __init__(self):
        super().__init__()
        self.data: Dict[str, Optional[str]] = {
            'login-croypto': None, 'login-page-flowkey': None, 'captcha-url': None}
        self._opened_tag: Optional[str] = None
        self._count = len(self.data)

    def parse(self, page: str) -> Dict[str, str]:
        try:
            self.feed(page)
        except self._AllValuesGot:
            pass
        assert not self._count
        return self.data  # type: ignore

    def handle_starttag(self, tag, attrs):
        if tag == 'p':
            for attr in attrs:
                if attr[0] == 'id':
                    if attr[1] in self.data:
                        self._opened_tag = attr[1]
                        assert self.data[attr[1]] is None
                    return

    def handle_data(self, data):
        if self._opened_tag is not None:
            self.data[self._opened_tag] = data.strip()
            self._opened_tag = None
            self._count -= 1
            if not self._count:
                raise self._AllValuesGot


class _SSOErrorParser(HTMLParser):
    class _ErrorGot(Exception):
        pass

    def __init__(self):
        super().__init__()
        self._error_code_str: str = ""
        self._error_div_opened: bool = False

    def parse(self, page: str) -> Optional[int]:
        try:
            self.feed(page)
        except self._ErrorGot:
            return int(self._error_code_str.strip())
        return None

    def handle_starttag(self, tag, attrs):
        if tag == 'div':
            if ("id", "login-error-msg") in attrs:
                self._error_div_opened = True

    def handle_endtag(self, tag):
        if self._error_div_opened and tag == 'div':
            raise self._ErrorGot

    def handle_data(self, data):
        if self._error_div_opened:
            self._error_code_str += data


class _LoginedPageParser(HTMLParser):
    MSG_ATTRS = [("id", "msg"), ("class", "login_auth_error")]
    KICK_TABLE_ATTRS = [("class", "kick_table")]
    KICK_POST_ATTRS = [('method', 'post'), ('id', 'continue')]
    CANCEL_POST_ATTRS = [('method', 'post'), ('id', 'cancel')]

    def __init__(self, status_code: int):
        super().__init__()
        self._msg: bool = False
        self._kick: bool = False
        self._waiting_kick_excution: bool = False
        self._kick_execution: str = ""
        self._waiting_cancel_excution: bool = False
        self._cancel_execution: str = ""
        self.status_code: int = status_code

    def handle_starttag(self, tag, attrs):
        if tag == "span" and attrs == self.MSG_ATTRS:
            self._msg = True
        elif tag == "table" and attrs == self.KICK_TABLE_ATTRS:
            self._kick = True
        elif tag == "form" and attrs == self.CANCEL_POST_ATTRS:
            self._waiting_cancel_excution = True
        elif tag == "form" and attrs == self.KICK_POST_ATTRS:
            self._waiting_kick_excution = True
        elif tag == "input" and ("name", "execution") in attrs:
            if self._waiting_kick_excution:
                for key, value in attrs:
                    if key == "value":
                        self._kick_execution = value
                        self._waiting_kick_excution = False
            elif self._waiting_cancel_excution:
                for key, value in attrs:
                    if key == "value":
                        self._cancel_execution = value
                        self._waiting_cancel_excution = False

    def handle_data(self, data):
        if self._msg:
            error_str = data.strip()
            if error_str == "无效的验证码":
                raise InvaildCaptcha()
            elif error_str == "您提供的用户名或者密码有误":
                raise IncorrectLoginCredentials()
            else:
                raise UnknownAuthserverException(
                    f"status code {self.status_code} is got (302 expected)"
                    f" when sending login post, {error_str}"
                )

def _random_str(length: int) -> str:
    return ''.join(random.choices(_CHAR_SET, k=length))

def _get_formdata(html: str, username: str, password: str) -> Dict[str, Optional[str]]:
    # from https://github.com/CQULHW/CQUQueryGrade
    parser = _AuthPageParser()
    parser.feed(html)
    salt = parser.salt
    if not salt:
        ParseError("无法获取盐")
    assert salt
    passwd_pkcs7 = pad16((_random_str(64)+str(password)).encode())
    encryptor = aes_cbc_encryptor(salt.encode(), _random_str(16).encode())
    passwd_encrypted = b64encode(encryptor(passwd_pkcs7)).decode()
    parser.input_data['username'] = username
    parser.input_data['password'] = passwd_encrypted
    return parser.input_data



from typing import Dict, Any

from pydantic import BaseModel

__all__ = ['RoomExamInvigilator']


class RoomExamInvigilator(BaseModel):
    """教室考试活动监考员信息"""
    name: str
    """姓名"""
    type: str
    """监考类型，如副监考、巡考等"""
    dept_name: str
    """所在部门名称"""

    @staticmethod
    def from_dict(data: Dict[str, Any]):
        """从反序列化的一个教室考试监考员信息 json 中生成RoomExamInvigilator对象

        :param data: 反序列化成字典的教室考试监考员信息 json
        :type data: Dict[str, Any]
        :return: 教室考试活动监考员对象
        :rtype: RoomExamInvigilator
        """
        return RoomExamInvigilator(
            name=data['name'],
            type=data['invigilatorType'],
            dept_name=data['deptName']
        )



from typing import Union, Optional, List, Dict, Any

from requests import Session
from pydantic import BaseModel

from ...course.models import CQUSession
from .room import Room
from .room_course import RoomCourse
from .room_exam import RoomExam
from .room_temp_activity import RoomTempActivity
from ..tools import get_room_timetable_raw, async_get_room_timetable_raw
from ...utils.request_transformer import Request

__all__ = ['RoomTimetable']


class RoomTimetable(BaseModel):
    course_timetable: List[RoomCourse]
    exam_timetable: List[RoomExam]
    temp_activity_timetable: List[RoomTempActivity]

    @staticmethod
    def from_dict(data: Dict[str, Any]):
        """从反序列化的一个教室信息 json 中生成RoomTimetable对象

        :param data: 反序列化成字典的教室信息 json
        :type data: Dict[str, Any]
        :return: 教室活动信息
        :rtype: RoomTimetable
        """
        return RoomTimetable(
            course_timetable=[RoomCourse.from_dict(temp) for temp in data['classTimetableVOList']] \
            if data['classTimetableVOList'] is not None else [],
            exam_timetable=[RoomExam.from_dict(temp) for temp in data['roomExamTimeTableVOList']] \
            if data['roomExamTimeTableVOList'] is not None else [],
            temp_activity_timetable=[RoomTempActivity.from_dict(temp) for temp in data['tempActivityTimetableVOList']] \
            if data['tempActivityTimetableVOList'] is not None else []
        )

    @staticmethod
    def fetch(session: Session, room: Union[Room, str], cqu_session: Optional[Union[CQUSession, str]] = None):
        """
        获取某教室活动详情

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :param room: 教室信息（为Room对象或需要获取的教室名称）
        :type room: Union[Room, str]
        :param cqu_session: 需要获取课表的学期，留空获取当前年级的课表
        :type cqu_session: Optional[Union[CQUSession, str]], optional
        :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
        :raises InvalidName: 若教室名称不为准确教室名称时
        :return: 教室活动信息对象
        :rtype: RoomTimetable
        """

        return RoomTimetable.from_dict(get_room_timetable_raw(session, room, cqu_session))

    @staticmethod
    async def async_fetch(session: Request, room: Union[Room, str], cqu_session: Optional[Union[CQUSession, str]] = None):
        """
        获取某教室活动详情

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :param room: 教室信息（为Room对象或需要获取的教室名称）
        :type room: Union[Room, str]
        :param cqu_session: 需要获取课表的学期，留空获取当前年级的课表
        :type cqu_session: Optional[Union[CQUSession, str]], optional
        :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
        :raises InvalidName: 若教室名称不为准确教室名称时
        :return: 教室活动信息对象
        :rtype: RoomTimetable
        """
        return RoomTimetable.from_dict(await async_get_room_timetable_raw(session, room, cqu_session))



from __future__ import annotations

from typing import Any, Dict, List

from requests import Session
from pydantic import BaseModel

from ...exception import MycquUnauthorized
from ...utils.request_transformer import Request, RequestTransformer

ROOM_ID_URL = "https://my.cqu.edu.cn/api/resourceapi/room/roomName-filter"

__all__ = ['Room']


class Room(BaseModel):
    """教室对象，储存了某个教室的相关信息"""
    id: int
    """教室id"""
    name: str
    """教室名称，如D1345"""
    capacity: int
    """教室容量"""
    building_name: str
    """教室所属建筑名称"""
    campus_name: str
    """教室所属校区"""
    room_type: str
    """教室类型"""

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> Room:
        """从反序列化的一个教室信息 json 中生成教室对象

        :param data: 反序列化成字典的教室 json
        :type data: Dict[str, Any]
        :return: 教室对象
        :rtype: Room
        """
        return Room(
            id=int(data['id']),
            name=data['name'],
            capacity=int(data['capacity']),
            building_name=data['buildingName'],
            campus_name=data['campusName'],
            room_type=data['roomClassificationName']
        )

    @staticmethod
    @RequestTransformer.register()
    def _fetch(session: Session, name: str) -> List[Room]:
        res = yield session.get(ROOM_ID_URL, params={'roomName': name})
        if res.status_code == 401:
            raise MycquUnauthorized

        return [Room.from_dict(room) for room in res.json()]

    @staticmethod
    def fetch(session: Session, name: str) -> List[Room]:
        """
        依据教室名字查询教室（支持模糊查询）

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :param name: 教室名称
        :type name: str
        :return: 教室对象组成的列表
        :rtype: List[Room]
        """
        return Room._fetch.sync_request(session, name)

    @staticmethod
    async def async_fetch(session: Request, name: str) -> List[Room]:
        """
        依据教室名字查询教室（支持模糊查询）

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :param name: 教室名称
        :type name: str
        :return: 教室对象组成的列表
        :rtype: List[Room]
        """
        return await Room._fetch.async_request(session, name)




"""教室相关信息模块"""

from __future__ import annotations
from typing import Optional, Union

from requests import Session
from ..course import CQUSession, CQUSessionInfo
from .models.room import Room
from ..exception import MycquUnauthorized, InvalidRoom
from ..utils.request_transformer import Request, RequestTransformer

ROOM_TIMETABLE_URL = "https://my.cqu.edu.cn/api/timetable/class/timetable/room/table-detail?sessionId=1039"

__all__ = ['get_room_timetable_raw', 'async_get_room_timetable_raw']


@RequestTransformer.register()
def _get_room_timetable_raw(session: Session, room: Union[Room, str],
                            cqu_session: Optional[Union[CQUSession, str]] = None):
    if cqu_session is None:
        cqu_session = (yield CQUSessionInfo._fetch).session
    elif isinstance(cqu_session, str):
        cqu_session = CQUSession.from_str(cqu_session)
    assert isinstance(cqu_session, CQUSession)

    if isinstance(room, str):
        temp = yield Room._fetch, {'name': room}
        if len(temp) == 0 or temp[0].name != room:
            raise InvalidRoom
        else:
            room = temp[0]
    assert isinstance(room, Room)

    res = yield session.post(ROOM_TIMETABLE_URL, json=[str(room.id)])
    if res.status_code == 401:
        raise MycquUnauthorized

    return res.json()

def get_room_timetable_raw(session: Session, room: Union[Room, str],
                           cqu_session: Optional[Union[CQUSession, str]] = None):
    """
    获取某教室活动详情

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
    :type session: Session
    :param room: 教室信息（为Room对象或需要获取的教室名称）
    :type room: Union[Room, str]
    :param cqu_session: 需要获取课表的学期，留空获取当前年级的课表
    :type cqu_session: Optional[Union[CQUSession, str]], optional
    :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
    :raises InvalidName: 若教室名称不为准确教室名称时
    :return: 反序列化获取教室活动的json
    :rtype: dict
    """
    return _get_room_timetable_raw.sync_request(session, room, cqu_session)

async def async_get_room_timetable_raw(session: Request, room: Union[Room, str],
                                       cqu_session: Optional[Union[CQUSession, str]] = None):
    """
    异步的获取某教室活动详情

    :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
    :type session: Session
    :param room: 教室信息（为Room对象或需要获取的教室名称）
    :type room: Union[Room, str]
    :param cqu_session: 需要获取课表的学期，留空获取当前年级的课表
    :type cqu_session: Optional[Union[CQUSession, str]], optional
    :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
    :raises InvalidName: 若教室名称不为准确教室名称时
    :return: 反序列化获取教室活动的json
    :rtype: dict
    """
    return await _get_room_timetable_raw.async_request(session, room, cqu_session)



from .models import RequestProtocol, ResponseProtocol, Request, Response, Requestable
from .params_mapper import *
from .request_transformer import RequestTransformer

__all__ = [
    'Request', 'Response', 'Requestable',
    'ResponseProtocol', 'RequestProtocol', 'RequestTransformer',
    'RequestParamsMapper', 'RequestsParamsMapper', 'HttpxParamsMapper'
]



from typing import Dict

from .request_transformer.params_mapper import RequestsParamsMapper, HttpxParamsMapper

__all__ = ['ConfigManager']


PYMYCQU_CONFIG = {
    'request': {
        'sync_request_params_mapper': RequestsParamsMapper,
        'async_request_params_mapper': HttpxParamsMapper
    }
}


def singleton(cls):
    _instance = {}

    def inner() -> 'ConfigManager':
        if cls not in _instance:
            _instance[cls] = cls()
        return _instance[cls]
    return inner

@singleton
class ConfigManager:
    def __init__(self):
        self._config = PYMYCQU_CONFIG

    @property
    def config(self) -> Dict:
        return self._config

    @config.setter
    def config(self, value):
        self._config = value

